#!/usr/bin/env python
from Annotation import Annotation,tokenize_sv
from sv2_backend import format_chrom
from collections import OrderedDict
from pybedtools import BedTool
import sys
class VCF():
	def __init__(self,fh=None):
		self.fh=fh
		self.Annotations=None # Annotation object
		self.genotypes={} # [(c,s,e,cl)]=[genotypes]
		self.allele_freq={} # [locus]=[#alt alleles, total alleles]
		self.quals={} # [locus]=[ref,alt]
		self.filters={} # [locus]=[std,dnm]
		self.head=None
	def init_header(self,date=None,ids=None,Structural_Variant=None,gen=None):
		chroms,refs,contigs={},OrderedDict(),[]
		from sv2Config import Config
		with open('{}{}.genome'.format(Config().resource_path(),gen),'r') as f:
			for l in f:
				chrom, leng = l.rstrip().split('\t')
				chroms[format_chrom(chrom)]=leng
		for x in Structural_Variant.raw: 
			if chroms.get(format_chrom(x[0]))!=None: refs[x[0]]=chroms[format_chrom(x[0])]
		for chrom in refs: contigs.append('##contig=<ID={},length={}>'.format(chrom,refs[chrom]))
		self.head=[
			'##fileformat=VCFv4.1',
			'##fileDate={}'.format(date),
			'##SV2_CMD="{}"'.format(' '.join(map(str,sys.argv[:]))),
			'##INFO=<ID=END,Number=1,Type=Integer,Description="End position of structural variant">',
			'##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variant">',
			'##INFO=<ID=SVLEN,Number=1,Type=Integer,Description="Difference in length between REF and ALT alleles">',
			'##INFO=<ID=DENOVO_FILTER,Number=1,Type=String,Description="Stringent filter status, recommended for de novo mutation discovery">',
			'##INFO=<ID=REF_GTL,Number=1,Type=Float,Description="Median Phred-adjusted REF genotype likelihood">',
			'##INFO=<ID=AF,Number=1,Type=Float,Description="Alternate allele frequency,in the range (0,1)">',
			'##INFO=<ID=CYTOBAND,Number=.,Type=String,Description="Cytoband(s) overlapping the variant">',
			'##INFO=<ID=REPEATMASKER,Number=2,Type=String,Description="Name and reciprocal overlap of RepeatMasker variant">',
			'##INFO=<ID=1000G_ID,Number=1,Type=String,Description="1000 Genomes Phase 3 integrated SV callset variant identifier">',
			'##INFO=<ID=1000G_OVERLAP,Number=1,Type=Float,Description="Overlap to 1000 Genomes Phase 3 variant, in the range (0,1)">',
			'##INFO=<ID=DESCRIPTION,Number=1,Type=String,Description="Verbose description of SV, 1-based coordinates"',
			'##INFO=<ID=GENES,Number=1,Type=String,Description="Genes overlapping the variant, pipe-separated by transcripts>"',
			'##INFO=<ID=ABPARTS,Number=1,Type=Float,Description="Overlap to antibody parts, in the range (0,1)">',
			'##INFO=<ID=CENTROMERE,Number=1,Type=Float,Description="Centromere overlap, in the range (0,1)">',
			'##INFO=<ID=GAP,Number=1,Type=Float,Description="Overlap to gaps in the reference, in the range (0,1)">',
			'##INFO=<ID=SEGDUP,Number=1,Type=Float,Description="Segmental duplication overlap, in the range (0,1)">',
			'##INFO=<ID=STR,Number=1,Type=Float,Description="Short tandem repeat overlap, in the range (0,1)">',
			'##INFO=<ID=UNMAPPABLE,Number=1,Type=Float,Description="Overlap to DAC Blacklisted Regions, in the range (0,1)">',
			'##FILTER=<ID=ABPARTS,Description="Variant overlaps to antibody parts >50%">',
			'##FILTER=<ID=CENTROMERE,Description="Variant overlaps to centromere >50%">',
			'##FILTER=<ID=GAP>,Description="Variant overlaps to reference gaps >50%">',
			'##FILTER=<ID=GENOTYPEFAIL,Description="Variant was unable to be genotyped">',
			'##FILTER=<ID=NOALT,Description="No alternate allele detected">',
			'##FILTER=<ID=SEGDUP,Description="Variant overlaps to segmental duplications >50%">',
			'##FILTER=<ID=STR,Description="Variant overlaps to short tandem repeats >50%">',
			'##FILTER=<ID=UNMAPPABLE,Description="Variant overlaps to DAC Blacklisted Regions regions >50%">',
			'##FILTER=<ID=FAIL,Description="Variant failed standard filters">',
			'##FILTER=<ID=PASS,Description="Variant passed standard filters">',
			'##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
			'##FORMAT=<ID=CN,Number=1,Type=Float,Description="Copy number estimate">',
			'##FORMAT=<ID=PE,Number=1,Type=Float,Description="Normalized discordant paired-end count">',
			'##FORMAT=<ID=SR,Number=1,Type=Float,Description="Normalized split-read count">',
			'##FORMAT=<ID=SC,Number=1,Type=Float,Description="SNV normalized coverage">',
			'##FORMAT=<ID=NS,Number=1,Type=Integer,Description="Number of SNVs within locus">',
			'##FORMAT=<ID=HA,Number=1,Type=Float,Description="Heterozygous allele ratio">',
			'##FORMAT=<ID=NH,Number=1,Type=Integer,Description="Number of heterozygous SNVs">',
			'##FORMAT=<ID=SQ,Number=1,Type=Float,Description="Phred-scaled genotype likelihood">',
			'##FORMAT=<ID=GL,Number=2|3,Type=Float,Description="Phred-scaled genotype likelihoods in the order, REF:(0/0), HET:(0/1), HOM:(1/1)">',
			'##ALT=<ID=DEL,Description="Deletion, if 80% reciprocal overlap with RepeatMasker element, the class, name, and family are given separated by colons">',
			'##ALT=<ID=DUP,Description="Duplication, if 80% reciprocal overlap with RepeatMasker element, the class, name, and family are given separated by colons">',
			'{}'.format('\n'.join(contigs)),
			'#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t{}'.format('\t'.join(ids)),
		]
	def init_row(self,data=None):
		svtype, _1kgp_id, _1kgp_ovr,mei_id,mei_ovr= str(data[3]),'NA','NA','NA',0
		gene,cytoband,median_ref,qual,allele,std_flt,dnm_flt = 'intergenic','NA','NA','NA','NA','PASS','NA'
		locus = tokenize_sv(data)
		svlen=locus[2]-locus[1]
		if self.quals.get(locus+(svtype,))!=None: median_ref,qual = self.quals[locus+(svtype,)]
		if self.allele_freq.get(locus+(svtype,))!=None: allele=float(format(float(self.allele_freq[locus+(svtype,)][0])/float(self.allele_freq[locus+(svtype,)][1]),'.3f'))
		if self.filters.get(locus+(svtype,))!=None: std_flt,dnm_flt=self.filters[locus+(svtype,)]
		description='{}:{}-{}_{}'.format(data[0],int(data[1])+1,data[2],svtype.replace('DEL','deletion').replace('DUP','duplication'))
		fail=[]
		if self.Annotations!=None:
			if self.Annotations.repeatmasker.get(locus)!=None:
				mei_id,mei_ovr= self.Annotations.repeatmasker[locus]
				mei_ovr= format(mei_ovr,'.2f')
				svtype=svtype+':'+mei_id
			if self.Annotations.cytoband.get(locus)!=None: cytoband=self.Annotations.cytoband[locus]
			abpart_ovr = check_dict(self.Annotations.excluded,locus+('abparts',))
			centromere_ovr = check_dict(self.Annotations.excluded,locus+('centromere',))
			gap_ovr = check_dict(self.Annotations.excluded,locus+('gap',))
			segdup_ovr = check_dict(self.Annotations.excluded,locus+('segDup',))
			str_ovr = check_dict(self.Annotations.excluded,locus+('STR',))
			unmap_ovr = check_dict(self.Annotations.excluded,locus+('unmappable',))
			if self.Annotations._1kgp.get(locus+(str(data[3]),))!=None: _1kgp_id,_1kgp_ovr=self.Annotations._1kgp[locus+(str(data[3]),)]
			if self.Annotations.genes.get(locus)!=None: gene=self.Annotations.genes[locus]
			# filters
			if float(abpart_ovr) >= 0.5: fail.append('ABPARTS')
			if float(centromere_ovr) >= 0.5: fail.append('CENTROMERE')
			if float(gap_ovr)>=0.5: fail.append('GAP')
			if float(segdup_ovr) >= 0.5: fail.append('SEGDUP')
			if float(str_ovr) >= 0.5: fail.append('STR')
			if float(unmap_ovr) >= 0.5: fail.append('UNMAPPABLE')
		if median_ref=='NA' and qual=='NA': fail.append('GENOTYPEFAIL')
		if allele==0: fail.append('NOALT')
		if std_flt=='FAIL': fail.append('FAIL')
		elif std_flt=='PASS' and ('GENOTYPEFAIL' not in fail and 'NOALT' not in fail): fail.append('PASS')
		if len(fail) > 0: filt = ','.join(fail)
		if std_flt=='FAIL':dnm_flt='FAIL'
		if ('NOALT' in filt or 'GENOTYPEFAIL' in filt) and dnm_flt=='PASS':dnm_flt='NA' 
		if self.Annotations!=None:
			info = 'END={};SVTYPE={};SVLEN={};DENOVO_FILTER={};REF_GTL={};AF={};CYTOBAND={};REPEATMASKER={},{};1000G_ID={};1000G_OVERLAP={};DESCRIPTION={};GENES={};ABPARTS={};CENTROMERE={};GAP={};SEGDUP={};STR={};UNMAPPABLE={}'.format(data[2],svtype,svlen,dnm_flt,median_ref,allele,cytoband,mei_id,mei_ovr,_1kgp_id,_1kgp_ovr,description,gene,abpart_ovr,centromere_ovr,gap_ovr,segdup_ovr,str_ovr,unmap_ovr)
		else:
			info = 'END={};SVTYPE={};SVLEN={};DENOVO_FILTER={};REF_GTL={};AF={};'.format(data[2],svtype,svlen,dnm_flt,median_ref,allele)
		return '<{}>\t{}\t{}\t{}\tGT:CN:PE:SR:SC:NS:HA:NH:SQ:GL'.format(svtype,qual,filt,info)
	def load_genotypes(self,Structural_Variant=None,SVs=None,Ped=None,ids=None,gen=None,no_anno=None,tmp_dir=None):
		svs=BedTool(list(set([(format_chrom(x[0]),x[1],x[2]) for x in Structural_Variant.raw]))).sort()
		if no_anno==False:
			Annot = Annotation()
			Annot.check_overlap(svs,Structural_Variant.raw,gen,tmp_dir)
			self.Annotations=Annot
		for locus in SVs:
			Variant = SVs[locus]
			self.quals[locus]=Variant.med_ref,Variant.med_alt
			self.filters[locus]=Variant.standard_filter,Variant.denovo_filter
			for sample_id in ids:
				gt='./.'
				if (locus[0]=='chrX' or locus[0]=='chrY') and Ped.males.get(sample_id)!=None and Structural_Variant.par[locus]==False: gt='.'
				if Variant.gt.get(locus+(sample_id,))!=None:
					gt=Variant.gt[locus+(sample_id,)]
					for allele in gt.split(':').pop(0).split('/'):
						if allele=='.': continue
						if self.allele_freq.get(locus)==None: self.allele_freq[locus]=[int(allele),1]
						else: self.allele_freq[locus]=[self.allele_freq[locus][0]+int(allele),self.allele_freq[locus][1]+1]
				if self.genotypes.get(locus)==None: self.genotypes[locus]=[gt]
				else: self.genotypes[locus].append(gt)
def check_dict(d,key):
	if d.get(key)!=None: return float(format(d[key],'.3f'))
	else: return 0


#!/usr/bin/env python2
import pandas as pd
import numpy as np
try: import cPickle as pickle
except ImportError: import pickle
def biallelic_del_svm(df,gtclf):
	gt_df,gt_mat,lt_df,lt_mat = prep_biallelic_dels(df)
	final=pd.DataFrame()
	if len(gt_mat) > 0:
		clf=None
		with open(gtclf['delgt'],'rb') as f: clf=pickle.load(f)
		final=final.append(biallelic_del_tabulate(gt_df,clf.predict_proba(gt_mat),'DEL_gt1kb'))
	if len(lt_mat) > 0:
		clf=None
		with open(gtclf['dellt'],'rb') as f: clf=pickle.load(f)
		final=final.append(biallelic_del_tabulate(lt_df,clf.predict_proba(lt_mat),'DEL_lt1kb'))
	return final
def biallelic_dup_svm(df,gtclf):
	snv_df,brk_df = prep_biallelic_dups(df)
	final=pd.DataFrame()
	if len(snv_df) >0:
		clf=None
		snv_mat = tabulate_snv(snv_df)
		with open(gtclf['duphar'],'rb') as f: clf=pickle.load(f)
		final=final.append(biallelic_dup_tabulate(snv_df,clf.predict_proba(snv_mat),'DUP_har'))
	if len(brk_df) >0:
		clf=None
		brk_mat= tabulate_3feats(brk_df)
		with open(gtclf['dupbrk'],'rb') as f: clf=pickle.load(f)
		final=final.append(biallelic_dup_tabulate(brk_df,clf.predict_proba(brk_mat),'DUP_breakpoint'))
	return final
def male_sex_chrom_svm(df,gtclf,clfname):
	df[['covr_GC','dpe','sr']]=df[['covr_GC','dpe','sr']].astype(float)
	df['covr']=df['covr_GC']
	df = df[df['covr'] < 5.0]
	mat =  tabulate_3feats(df)
	final=pd.DataFrame()
	if len(mat)>0:
		clf=None
		with open(gtclf[clfname],'rb') as f: clf=pickle.load(f)
		if clfname=='delmsc':
			final=final.append(male_sex_chrom_del_tabulate(df,clf.predict_proba(mat),'DEL_male_sex_chrom'))
		else:
			final=final.append(male_sex_chrom_dup_tabulate(df,clf.predict_proba(mat),'DUP_male_sex_chrom'))
	return final
def training_features(df,case,out): # case 0:dels; 1:dups; 2:del_msc; 3:dup_msc
	pd.options.mode.chained_assignment = None
	if case==0:
		gt_df,gt_mat,lt_df,lt_mat = prep_biallelic_dels(df)
		if len(gt_df)>0:
			ofh = ofh_train('{}_deletion_gt1kb.txt'.format(out))
			for x in train_reformat(gt_df).values: ofh.write('\t'.join(map(str,x))+'\tNA\tdeletion_gt1kb\n')
			ofh.close()
		if len(lt_df)>0:
			ofh = ofh_train('{}_deletion_lt1kb.txt'.format(out))
			for x in train_reformat(lt_df).values: ofh.write('\t'.join(map(str,x))+'\tNA\tdeletion_lt1kb\n')
	if case==1:
		snv_df,brk_df = prep_biallelic_dups(df)
		if len(brk_df)>0:
			ofh = ofh_train('{}_duplication_breakpoint.txt'.format(out))
			for x in train_reformat(brk_df).values: ofh.write('\t'.join(map(str,x))+'\tNA\tduplication_breakpoint\n')
			ofh.close()
		if len(snv_df)>0:
			ofh= ofh_train('{}_duplication_har.txt'.format(out))
			for x in train_reformat(snv_df).values: ofh.write('\t'.join(map(str,x))+'\tNA\tduplication_har\n')
			ofh.close()
	if case==2 or case==3:
		df[['covr_GC','dpe','sr']]=df[['covr_GC','dpe','sr']].astype(float)
		df['covr']=df['covr_GC']
		df = df[df['covr'] < 5.0]
		out_suffix = 'deletion_male_sex_chrom'
		if case==3: out_suffix='duplication_male_sex_chrom'
		if len(df)>0:
			ofh = ofh_train('{}_{}.txt'.format(out,out_suffix))
			for x in train_reformat(df).values: ofh.write('\t'.join(map(str,x))+'\tNA\t{}\n'.format(out_suffix))
			ofh.close()
cdef data_table(df,lik,cn,biallelic=True):
	df['copy_number']=cn
	if biallelic==True:
		df['likHOM'],df['likHET'],df['likREF']=lik[:,0],lik[:,1],lik[:,2]
	else:
		df['likHOM'],df['likHET'],df['likREF']=lik[:,0],lik[:,0],lik[:,1]
	df['NONREF'] = 1 - df['likREF']
	pref = 1.0 - df['likREF']
	palt = 1.0 - df['NONREF']
	pref[pref==1]=1-1e-12
	pref[pref==0]=0+1e-12
	palt[palt==1]=1-1e-12
	palt[palt==0]=0+1e-12
	df['PHRED_REF'] = -10.0 * np.log10(pref)
	df['PHRED_NONREF'] = -10.0 * np.log10(palt)
	df= df[['chr','start','end','type','size','id','covr','dpe','sr','SNP_coverage','SNPs','HET_ratio','HET_SNPs','copy_number','likREF','likHET','likHOM','NONREF','PHRED_REF','PHRED_NONREF','classif']]	
	return df
cdef biallelic_del_tabulate(df,lik,clfname):
	df['classif']=clfname
	cn=[]
	for x in lik:
		if x[0] > x[1] and x[0] > x[2]: cn.append(0)
		elif x[1] > x[0] and x[1] > x[2]: cn.append(1)
		elif x[2] > x[0] and x[2] > x[1]: cn.append(2)
		else: cn.append(2)
	return data_table(df,lik,cn)
cdef biallelic_dup_tabulate(df,lik,clfname):
	df['classif']=clfname
	cn=[]
	for x in lik:
		if x[0] > x[1] and x[0] > x[2]: cn.append(4)
		elif x[1] > x[0] and x[1] > x[2]: cn.append(3)
		elif x[2] > x[0] and x[2] > x[1]: cn.append(2)
		else: cn.append(2)	
	return data_table(df,lik,cn)
cdef male_sex_chrom_del_tabulate(df,lik,clfname):
	df['classif']=clfname
	cn=[]
	for x in lik:
		if x[0] > x[1]: cn.append(0)
		else: cn.append(1)
	return data_table(df,lik,cn,False)
cdef male_sex_chrom_dup_tabulate(df,lik,clfname):
	df['classif']=clfname
	cn=[]
	for x in lik:
		if x[0] > x[1]: cn.append(2)
		else: cn.append(1)
	return data_table(df,lik,cn,False)
cdef ofh_train(o):
	ofh = open(o,'w')
	ofh.write('\t'.join(('chr','start','end','type','size','id','covr','dpe_rat','sr_rat','SNP_coverage','SNPs','HET_ratio','HET_SNPs','copy_number','classifier'))+'\n')
	return ofh
cdef prep_biallelic_dels(df):
	df[['covr_GC','dpe','sr']]=df[['covr_GC','dpe','sr']].astype(float)
	df[['size']]=df[['size']].astype(int)
	df = df[df['covr_GC'] < 5.0]
	df['covr']=df['covr_GC']
	gt_df = df[df['size'] > 1000]
	lt_df = df[df['size'] <= 1000]
	gt_mat = tabulate_3feats(gt_df)
	lt_mat = tabulate_3feats(lt_df)
	return gt_df,gt_mat,lt_df,lt_mat
cdef prep_biallelic_dups(df):
	snv_df,brk_df= pd.DataFrame(),pd.DataFrame()
	df[['SNPs','HET_SNPs']]=df[['SNPs','HET_SNPs']].astype(int)
	df[['covr_GC','dpe','sr']]=df[['covr_GC','dpe','sr']].astype(float)
	mean_covr=[]
	for ind,ele in df.iterrows():
		if ele['SNPs'] > 0: mean_covr.append(np.mean((ele['covr_GC'],float(ele['SNP_coverage']))))
		else: mean_covr.append(ele['covr_GC'])
	df['covr']=mean_covr
	df=df[df['covr'] < 5.0]
	snvs=[]
	brks=[]
	for ind,ele in df.iterrows():
		if ele['dpe']==0 and ele['sr']==0 and ele['HET_SNPs'] > 0: snvs.append(ind)
		else: brks.append(ind)
	snv_df=df.loc[snvs]
	snv_df[['HET_ratio']]=snv_df[['HET_ratio']].astype(float)
	brk_df=df.loc[brks]
	return snv_df,brk_df
cdef tabulate_3feats(df): return df[['covr','dpe','sr']].as_matrix()
cdef tabulate_snv(df): return df[['covr','HET_ratio']].as_matrix()
cdef train_reformat(df): return df[['chr','start','end','type','size','id','covr','dpe','sr','SNP_coverage','SNPs','HET_ratio','HET_SNPs']]	


#!/usr/bin/env python2
from sv2_backend import dump_json,errFH,get_path,query_yes_no
try:from configparser import ConfigParser
except ImportError:from ConfigParser import ConfigParser
import json,os,sys,wget,zipfile
__memsize__='238M'
__resource_url__ = 'https://github.com/dantaki/SV2/releases/download/sv2v1.4.0/sv2_resources.zip'
class Config():
	def __init__(self):
		self.fh=get_path()+'/config/sv2.ini'
		self.json=get_path()+'/config/sv2_clf.json'
		self.fasta_chr_flag=False
		self.clfs=None
		self.gtclf={} # classifiers for genotyping
		self.gcfh=get_path()+'/resources/GC_content_reference.txt'
		self.gc_dict={}
		self.resource=None # directory to resource files
		if not os.path.isfile(self.fh):
			print 'FATAL ERROR: {} config file is missing! Please reinstall sv2: [$ pip uninstall sv2 -y && pip install sv2-VERSION.tar.gz]'.format(self.fh)
			sys.stderr.write('FATAL ERROR: {} config file is missing! Please reinstall sv2: [$ pip uninstall sv2 -y && pip install sv2-VERSION.tar.gz]\n'.format(self.fh))
			sys.exit(1)
		if not os.path.isfile(self.json):
			print 'FATAL ERROR: {} json file is missing! Please reinstall sv2: [$ pip uninstall sv2 -y && pip install sv2-VERSION.tar.gz]'.format(self.json)
			sys.stderr.write('FATAL ERROR: {} json file is missing! Please reinstall sv2: [$ pip uninstall sv2 -y && pip install sv2-VERSION.tar.gz]\n'.format(self.json)) 		
			sys.exit(1)
		if not os.path.isfile(self.gcfh):
			print 'FATAL ERROR: {} file is missing! Please reinstall sv2: [$ pip uninstall sv2 -y && pip install sv2-VERSION.tar.gz]'.format(self.gcfh)
			sys.stderr.write('FATAL ERROR: {} file is missing! Please reinstall sv2: [$ pip uninstall sv2 -y && pip install sv2-VERSION.tar.gz]\n'.format(self.gcfh))
			sys.exit(1)
		with open(self.json) as data_file: self.clfs = json.load(data_file)
	def check_config(self):
		conf=ConfigParser()
		conf.read(self.fh)
		if conf.get('RESOURCE_DIR','sv2_resource')==None or not os.path.isdir(conf.get('RESOURCE_DIR','sv2_resource')):
			print 'FATAL ERROR: sv2 resource files not found.\nDownload the resource files with [ $ sv2 -download ] github.com/dantaki/SV2/wiki/Installation#download-resources'
			sys.stderr.write('FATAL ERROR: sv2 resource files not found.\nDownload the resource files with [ $ sv2 -download ] github.com/dantaki/SV2/wiki/Installation#download-resources\n')
			sys.exit(1)
		if not os.path.isfile(conf.get('FASTA_PATHS','hg19')) and not os.path.isfile(conf.get('FASTA_PATHS','hg38')) and not os.path.isfile(conf.get('FASTA_PATHS','mm10')):
			print 'FATAL ERROR: No reference fasta specified.\nConfigure sv2 before genotyping [ $ sv2 -hg19 FASTA -hg38 FASTA -mm10 FASTA ] github.com/dantaki/SV2/wiki/Installation#configure'
			sys.stderr.write('FATAL ERROR: No reference fasta specified.\nConfigure sv2 before genotyping [ $ sv2 -hg19 FASTA -hg38 FASTA -mm10 FASTA ] github.com/dantaki/SV2/wiki/Installation#configure\n')
			sys.exit(1)
	def download_resource(self,):
		install_loc = get_path()+'/resources/'
		change_loc = query_yes_no('Install sv2 resource files to the default location? < {} >'.format(install_loc))
		if change_loc == False:
			sys.stdout.write('Enter sv2 resource file destination:\n')
			install_loc = raw_input()
			if not install_loc.endswith('/'): install_loc=install_loc+'/'
			if not os.path.isdir(install_loc):
				sys.stderr.write('FATAL ERROR: {} not a directory\n'.format(install_loc))
				sys.stdout.write('FATAL ERROR: {} not a directory\n'.format(install_loc))
				sys.exit(1)
		install_ok = query_yes_no('sv2 resource files will require {} of free space in {}. Proceed with installation?'.format(__memsize__,install_loc))	
		if install_ok==False: sys.exit(1)
		else:
			zip_file = install_loc+'sv2_resources.zip'
			wget.download(__resource_url__,zip_file)
			zip_fh= zipfile.ZipFile(zip_file)
			if zip_fh.testzip() is not None:
				sys.stdout.write('FATAL ERROR: {} corrupted. Try <sv2 -download> again\n'.format(zip_file))
				sys.stderr.write('FATAL ERROR: {} corrupted. Try <sv2 -download> again\n'.format(zip_file))
				sys.exit(1)
			zip_fh.extractall(install_loc)
			zip_fh.close()
			os.remove(zip_file)
			self.resource = install_loc
			self.write_config(None,None,None)
		sys.stdout.write('INSTALL COMPLETE\nRun < sv2 -hg19 FASTA [-hg38,-mm10] > to configure sv2\n')
	def fasta(self,gen=None):
		conf = ConfigParser()
		conf.read(self.fh)
		fasta_path=conf.get('FASTA_PATHS',gen)
		if str(fasta_path)=='None':
			print 'FATAL ERROR: {} fasta for reference {} not found. Please configure SV2\n[ $ sv2 -hg19 FASTA -hg38 FASTA -mm10 FASTA ] github.com/dantaki/SV2/wiki/Installation#configure'.format(fasta_path,gen)
			sys.stderr.write('FATAL ERROR: {} fasta for reference {} not found. Please configure SV2\n[ $ sv2 -hg19 FASTA -hg38 FASTA -mm10 FASTA ] github.com/dantaki/SV2/wiki/Installation#configure\n'.format(fasta_path,gen))
			sys.exit(1)
		with open(fasta_path,'r') as f:
			l = f.next()
			if not l.startswith('>'):
				print 'FATAL ERROR: {} headers must contain ">"'.format(fasta_path)
				sys.stderr.write('FATAL ERROR: {} headers must contain ">"\n'.format(fasta_path))
				sys.exit(1)
			if l.startswith('>chr'): self.fasta_chr_flag=True
		return fasta_path
	def gc_norm(self,pcrfree):
		with open(self.gcfh,'r') as f:
			for l in f:
				tag,GC,norm = l.rstrip('\n').split('\t')
				key = 'DOC'
				if 'READCOUNT' in tag: key ='RC'
				if pcrfree == True:
					if 'PCRFREE' in tag: self.gc_dict[(key,int(GC))]=float(norm)
				else:
					if 'PCRFREE' not in tag: self.gc_dict[(key,int(GC))]=float(norm)
	def get_clf(self,classifier_name=None):
		names=[x for x in self.clfs['default']]
		if self.clfs.get(classifier_name)==None:
			sys.stderr.write('WARNING: {} classifier name not found. Using default classifiers. Please check {}\n'.format(classifier_name,self.json))
			classifier_name='default'
		if self.clfs.get(classifier_name)!=None:
			for x in names:
				if self.clfs[classifier_name].get(x)==None:
					sys.stderr.write('WARNING {} {} classifier not found. Using default {} classifier. Please check {}\n'.format(classifier_name,x,x,self.json))
					self.gtclf[x]=self.clfs['default'][x]
				else:
					if not os.path.isfile(self.clfs[classifier_name][x]):
						sys.stderr.write('WARNING {} {} pickle file not found. Using default {} classifier. Please check {} and reload the classifier json file <sv2 -load-clf <clf.json>\n'.format(classifier_name,x,x,self.json))
					else: self.gtclf[x]=self.clfs[classifier_name][x]
		else:
			print 'FATAL ERROR: {} classifier not found. Please check {}. If the default classifiers are not found, reinstall sv2 <pip uninstall sv2 -y && pip install sv2-VERSION-.tar.gz'.format(classifier_name,self.json)
			sys.stderr.write('FATAL ERROR: {} classifier not found. Please check {}. If the default classifiers are not found, reinstall sv2 <pip uninstall sv2 -y && pip install sv2-VERSION-.tar.gz\n'.format(classifier_name,self.json))
			sys.exit(1)
	def load_clf(self,jsonfh=None):
		import shutil
		errFH(jsonfh)
		clfs={}
		with open(jsonfh) as f: clfs = json.load(f)
		realpaths=[]
		for name in clfs:
			print 'loading {} classifier ...'.format(name)
			clf_dir = get_path()+'/resources/training_sets/'+name+'/'
			if not os.path.exists(clf_dir): os.makedirs(clf_dir)		
			for x in clfs[name]:
				clffh = str(clfs[name][x])
				if not os.path.isfile(clffh): sys.stderr.write('WARNING: {} does not exist. Please check the paths in {}\n'.format(clffh,jsonfh))
				else:
					clfname = clffh.split('/').pop()
					newpath = clf_dir+clfname
					clfreplace=True
					if os.path.isfile(newpath): clfreplace= query_yes_no('WARNING: {} exists... Replace?'.format(newpath),'no')
					if clfreplace==True:
						if self.clfs.get(name)==None: self.clfs[name]={}
						self.clfs[name][x]=newpath
						realpaths.append(newpath)
						shutil.copyfile(clffh, newpath)
		for x in realpaths: print 'installed classifier {}'.format(x)
		print 'appending to {}... DO NOT ALTER {}'.format(self.json,self.json)
		dump_json(self.json,self.clfs)
	def resource_path(self):
		conf = ConfigParser()
		conf.read(self.fh)
		resource_loc = conf.get('RESOURCE_DIR','sv2_resource')
		if not resource_loc.endswith('/'): resource_loc=resource_loc+'/'
		if not os.path.isdir(resource_loc):
			sys.stdout('FATAL ERROR: {} not found. Install resource files < sv2 -download >\n'.format(resource_loc))
			sys.stderr('FATAL ERROR: {} not found. Install resource files < sv2 -download >\n'.format(resource_loc))
			sys.exit(1)
		return resource_loc
	def write_config(self,hg19=None,hg38=None,mm10=None):
		conf=ConfigParser()
		for x in self.clfs['default']:
			clf_fh = str(self.clfs['default'][x])
			if not os.path.isfile(clf_fh):
				realpath = get_path()+'/resources/training_sets/'+clf_fh
				if os.path.isfile(realpath): self.clfs['default'][x]=realpath
				else:
					print 'FATAL ERROR: {} pickle file not found. If this file is missing, reinstall sv2: pip uninstall sv2 -y && pip install sv2-VERSION.tar.gz'.format(realpath)
					sys.stderr.write('FATAL ERROR: {} pickle file not found. If this file is missing, reinstall sv2: pip uninstall sv2 -y && pip install sv2-VERSION.tar.gz\n'.format(realpath))
					sys.exit(1)
		dump_json(self.json,self.clfs)
		if not os.path.isfile(self.fh):
			conf_fh = open(self.fh,'w')
			conf.add_section('FASTA_PATHS')
			conf.set('FASTA_PATHS','hg19',hg19)
			conf.set('FASTA_PATHS','hg38',hg38)
			conf.set('FASTA_PATHS','mm10',mm10)
			conf.set('RESOURCE_DIR','sv2_resource',self.resource)
			conf.write(conf_fh)
			conf_fh.close()
		else:
			conf.read(self.fh)
			if hg19 != None:
				errFH(hg19)
				conf.set('FASTA_PATHS','hg19',hg19)
			if hg38 != None:
				errFH(hg38)
				conf.set('FASTA_PATHS','hg38',hg38)
			if mm10 != None:
				errFH(mm10)
				conf.set('FASTA_PATHS','mm10',mm10)
			if self.resource != None:
				conf.set('RESOURCE_DIR','sv2_resource',self.resource)
			with open(self.fh,'w') as conf_fh: conf.write(conf_fh)


#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
from sv2_backend import format_chrom,get_path,make_dir,mask_bed,match_chrom_prefix
from Snv import preprocess_snv
import numpy as np
import os,pysam,sys
from pybedtools import BedTool
cimport numpy as np
np.import_array()
class Preprocess():
	def __init__(self,pre=None):
		chr_cov,insert_size,mad,read_length,snv_cov={},{},{},{},{}
		with open(pre) as f:
			for l in f:
				if l.startswith('#'):continue
				r = l.rstrip('\n').split('\t')
				chr_cov[(str(r[0]),str(r[1]))]=float(r[2])
				snv_cov[(str(r[0]),str(r[1]))]=float(r[7])
				if r[1] == "GENOME":
					mad[r[0]]=float(r[5])
					insert_size[r[0]]=float(r[4])
					read_length[r[0]]=float(r[3])
		self.chr_cov=chr_cov
		self.insert_size=insert_size
		self.insert_mad=mad
		self.read_len=read_length
		self.snv_cov=snv_cov
def preprocess(Bam,ofh,seed,gen,tmp_dir):
	outfh = open(ofh,'w')
	outfh.write('\t'.join(("#id","chrom","coverage_median","read_length_median","insert_size_median","insert_size_mad","bam_bp_parsed","snv_depth_median","snv_parsed"))+'\n')
	outfh.close()
	sv2_preprocess(Bam,ofh,seed,gen,tmp_dir)
cdef MAD(a, double c=0.6745): return np.around(np.median(np.fabs(a-np.median(a))/c),decimals=3)
cdef normalize_chr_cov(double read_count, double chr_size, read_length): return np.around( (read_count/chr_size)*np.median(read_length),decimals=2)
cdef random_int(): return ''.join(map(str,np.random.randint(9,size=6)))
cdef sv2_preprocess(Bam,ofh,seed,gen,tmp_dir):
	cdef unsigned int chr_size
	cdef unsigned int read_count
	cdef unsigned int genome_size
	genome_cov,genome_read_length,genome_insert_size,genome_mad,genome_size=[],[],[],[],0
	tmp_bed = Bam.tmp_chrom_file(tmp_dir,False) # tmp file chrom 0 chrom_len
	snv_depth=preprocess_snv(Bam,mask_bed(tmp_bed,gen))
	os.remove(tmp_bed)
	out = open(ofh,'a')
	Itr = pysam.AlignmentFile(Bam.fh,'r{}'.format(Bam.char))  
	from sv2Config import Config
	for chrom in Bam.refs:
		tmp_genome = Bam.tmp_chrom_file(tmp_dir,True,chrom)
		chrom=format_chrom(chrom)
		chr_size,read_count,read_stats,read_length,insert_size=0,0,{},[],[]
		rand_bed = BedTool()
		rand_bed = rand_bed.random(l=100000,n=100,seed=seed,g=tmp_genome)
		rand_bed = rand_bed.subtract(BedTool('{}{}_excluded.bed.gz'.format(Config().resource_path(),gen))).sort().merge()
		for entry in rand_bed:
			entry=tuple(entry)
			c,s,e = str(entry[0]),int(entry[1]),int(entry[2])
			if Bam.chr_flag==False: c=c.replace('chr','')
			region = '{}:{}-{}'.format(c,s+1,e)
			chr_size+=e-s
			for Aln in Itr.fetch(region=region):
				if (Aln.is_proper_pair == False or Aln.is_qcfail== True or Aln.is_duplicate == True or Aln.mapq < 40 or read_stats.get(str(Aln.qname)+str(Aln.is_read1)) != None): continue
				read_stats[str(Aln.qname)+str(Aln.is_read1)] = ((Aln.qlen,abs(Aln.isize)))
		read_count = len(read_stats)
		if read_count == 0: out.write('\t'.join(map(str,(Bam.id,chrom,'0','0','0','0',chr_size,snv_depth[chrom][0],snv_depth[chrom][1])))+'\n')
		else :
			for name in read_stats:
				read_length.append(read_stats[name][0])
				insert_size.append(read_stats[name][1])
			norm = normalize_chr_cov(read_count,chr_size,np.array(read_length))
			out.write('\t'.join(map(str,(Bam.id,chrom,norm,np.median(read_length),np.median(insert_size),MAD(np.array(insert_size)),chr_size,snv_depth[chrom][0],snv_depth[chrom][1])))+'\n')
			genome_cov.append(norm)
			genome_read_length.append(np.median(read_length))
			genome_insert_size.append(np.median(insert_size))
			genome_mad.append(MAD(np.array(insert_size)))
			genome_size += chr_size
		os.remove(tmp_genome)
	out.write('\t'.join(map(str,(Bam.id,'GENOME',np.median(genome_cov),np.median(genome_read_length),np.median(genome_insert_size),np.median(genome_mad),genome_size,snv_depth['GENOME'][0],snv_depth['GENOME'][1])))+'\n')
	out.close()
	Itr.close()


#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
from sv2_backend import accepted_chrom,format_chrom,match_chrom_key,match_chrom_prefix
from sv2Config import Config
from Preprocess import Preprocess
from Snv import extract_snv_features
from SV import skip_sv
import os,pysam,sys
import numpy as np
cimport numpy as np
np.import_array()
from pybedtools import BedTool
def extract_feats(Bam,sv,prefh,out,gen,pcrfree,legacyM,Conf,tmp_dir):
	master_sv = filter_sv(sv,Bam,gen,tmp_dir)
	Conf.gc_norm(pcrfree)
	sv_gc=gc_sv(sv,master_sv,gen,Conf)
	Pre = Preprocess(prefh)
	snv_feats,het_feats=extract_snv_features(Bam,sv,master_sv,Pre,gen)
	Itr = pysam.AlignmentFile(Bam.fh,'r{}'.format(Bam.char))
	insert_size, insert_mad = Pre.insert_size[Bam.id],Pre.insert_mad[Bam.id]
	ofh = open(out,'w')
	ofh.write('\t'.join(('#chr','start','end','type','size','id','coverage','coverage_GCcorrected','discordant_ratio','split_ratio','snv_coverage','heterozygous_allele_ratio','snvs','het_snvs'))+'\n')
	for call in sv:
		(c,s,e,cl) = call
		if sv_gc.get(call)==None or master_sv.get(call)==None: continue
		snv_depth,snvs,had,hets=float('nan'),0,float('nan'),0
		if snv_feats.get(call)!=None: snv_depth,snvs=snv_feats[call]
		if het_feats.get(call)!=None: had,hets=het_feats[call]
		chrom_key = match_chrom_key(call,Bam,gen)
		if Pre.chr_cov.get((Bam.id,chrom_key))==None: continue
		GC_content = sv_gc[call]
		sv_span,flank_span,windows = master_sv[call]
		size = int(e)-int(s)
		ci = Pre.insert_size[Bam.id]+(5*Pre.insert_mad[Bam.id])
		discordant,split,concordant = discordant_split_sv(flank_span,Itr,size,ci,windows,Bam.chr_flag,legacyM)
		if size > 1000:
			"""read count coverage estimation"""
			(Aln_count,bp_span) = count_reads(sv_span,Itr,ci,Bam.chr_flag,legacyM)
			cov=float('nan')
			if Pre.chr_cov[(Bam.id,chrom_key)] != 0: cov = normalize_coverage(Aln_count,bp_span,Pre.chr_cov[(Bam.id,chrom_key)],Pre.read_len[Bam.id])
			gc_norm_factor = 1.0
			if Conf.gc_dict.get(('RC',GC_content))==None:
				if Conf.gc_dict.get(('DOC',GC_content)) !=None: gc_norm_factor=Conf.gc_dict[('DOC',GC_content)]
			else: gc_norm_factor = Conf.gc_dict[('RC',GC_content)]
			gc_cov = cov * gc_norm_factor
		else:
			"""median depth of coverage estimation"""
			median_doc = depth_of_coverage(sv_span,Bam.fh,Bam.chr_flag,Pre.read_len[Bam.id])
			cov=float('nan')
			if Pre.chr_cov[(Bam.id,chrom_key)]!= 0: cov = median_doc/Pre.chr_cov[(Bam.id,chrom_key)]
			gc_norm_factor=1.0
			if Conf.gc_dict.get(('DOC',GC_content)) != None: gc_norm_factor=Conf.gc_dict[('DOC',GC_content)]
			gc_cov = cov * gc_norm_factor
		if float(concordant) == 0.0:
			discordant_ratio = str(round(float(discordant)/1.0,3))
			split_ratio = str(round(float(split)/1.0,3))
		else:
			discordant_ratio = str(round(float(discordant)/float(concordant),3))
			split_ratio = str(round(float(split)/float(concordant),3))
		ofh.write('\t'.join(map(str,(c,s,e,cl,size,Bam.id,round(float(cov),3),round(float(gc_cov),3),discordant_ratio,split_ratio,round(float(snv_depth),3),round(float(had),3),snvs,hets)))+'\n')
	Itr.close()
	ofh.close()
cdef append_list2dict(a):
	d={}
	for i in a:
		(c,s,e,cl,tag) = i
		if d.get(tag) == None: d[tag]=[(c,s,e)]
		else: d[tag].append((c,s,e))
	return d
cdef count_reads(sv_list,Itr,ci,chr_flag,legacyM):
	"""count reads within SV span for coverage estimation"""
	cdef unsigned int Aln_count
	cdef unsigned int bp_span
	Aln_count,bp_span=0,0
	for (c,s,e) in sv_list:
		c =  match_chrom_prefix(c,chr_flag)
		region= str('{}:{}-{}').format(c,int(s)+1,e)
		bp_span+=int(e)-int(s)
		"""count each Aln within the span"""
		for Aln in Itr.fetch(region=region):
			if (Aln.is_reverse == Aln.mate_is_reverse or Aln.is_proper_pair == False or Aln.is_qcfail == True or Aln.mapping_quality < 10
			   or ( (Aln.is_secondary and legacyM == True) or (Aln.is_supplementary and legacyM == False) )
			   or Aln.is_unmapped or Aln.mate_is_unmapped or Aln.is_duplicate or abs(Aln.tlen) >= ci or Aln.tid != Aln.rnext):
				continue
			Aln_count+=1
	return(Aln_count,bp_span)
cdef depth_of_coverage(sv_list,bamfh,chr_flag,Aln_length):
	"""return median depth of coverage for SV <= 1kb"""
	pos_doc={}
	for (c,s,e) in sv_list:
		c =  match_chrom_prefix(c,chr_flag)
		region= str('{}:{}-{}').format(c,int(s)+1,e)
		depth_result = pysam.depth("-a", "-Q" "40", "-r", region, "-l", str(Aln_length-10), bamfh)
		str_flag=0
		if isinstance(depth_result,str):
			depth_result = depth_result.split('\n')
			str_flag=1
		for x in depth_result:
			r = x.rstrip('\n').split('\t')
			if str_flag == 1:
				if len(r)!=3: continue
				pos_doc[int(r[1])]=int(r[2])
			else: pos_doc[int(r[1])]=int(r[2])
	if len(pos_doc) !=0:
		temp=[]
		for x in pos_doc: temp.append(float(pos_doc[x]))
		return np.median(temp)
	else: return 0
cdef discordant_split_sv(flank_list,Itr,size,ci,windows,chr_flag,legacyM):
	"""count discordant paired ends and split Alns"""
	cdef unsigned int discordant_count
	cdef unsigned int split_count
	cdef unsigned int concordant_count	
	discordant_count,split_count,concordant_count=0,0,0
	if flank_list==None: return (0,0,0)
	else:
		for (c,s,e) in flank_list:
			c =  match_chrom_prefix(c,chr_flag)
			region= str('{}:{}-{}').format(c,int(s)+1,e)
			for Aln in Itr.fetch(region=region,until_eof=True):
				if (Aln.is_qcfail or Aln.is_unmapped or Aln.mate_is_unmapped or Aln.tid != Aln.rnext or Aln.is_reverse == Aln.mate_is_reverse or Aln.is_duplicate): continue
				else:
					mate_pos = Aln.pnext
					if is_discordant(Aln,windows,ci,mate_pos) == True: discordant_count+=1
					if is_split(Aln,windows,c,legacyM)  == True: split_count+=1
					if (Aln.is_proper_pair and Aln.mapping_quality >= 10 and ( (not Aln.is_secondary and legacyM == True) or (not Aln.is_supplementary and legacyM == False) ) and abs(Aln.tlen) < ci):
						concordant_count+=1
					else: continue
		return(discordant_count,split_count,concordant_count)
cdef filter_sv(sv,Bam,gen,tmp_dir):
	#returns master_sv[sv_dict[int(i)]]=([sv_span],[flank_span,windows])
	cdef unsigned int tag=0
	annot_sv, sv_dict, start_flank, end_flank=[], {}, [],[]
	for r in sv:
		sv_dict[tag]=r
		(c,s,e,cl) = r
		chrom=format_chrom(c)
		c = match_chrom_prefix(c,Bam.chr_flag) # chromosome matches ones in the BAM file
		if c not in accepted_chrom(Bam.chr_flag,gen):
			sys.stderr.write('WARNING: Skipping {}. SV2 does not genotype SVs on contig {}\n'.format(r,c))
			continue
		if skip_sv(c,s,e,cl,Bam.refs[c])==True: continue
		annot_sv.append((chrom,s,e,cl,tag))
		start_flank.append((chrom,s,int(s)+1,cl,tag))
		end_flank.append((chrom,e,int(e)+1,cl,tag))
		tag+=1
	"""add 500bp to each start and end position"""
	_prefix='chr'
	if Bam.chr_flag==True: _prefix='' # if reference lengths are prefixed with chr, pass an empty string
	tmp_genome = tmp_dir+'/{}.genome'.format(Bam.id)
	tmpfh = open(tmp_genome,'w')
	for ref in Bam.refs: tmpfh.write('{}{}\t{}\n'.format(_prefix,ref,Bam.refs[ref]))
	tmpfh.close()
	start_flank500bp=BedTool(start_flank).slop(b=500,g=tmp_genome)
	end_flank500bp=BedTool(end_flank).slop(b=500,g=tmp_genome)
	start_flank_dict=dict((i[4],(i[1],i[2])) for i in start_flank500bp)
	end_flank_dict=dict((i[4],(i[1],i[2])) for i in end_flank500bp)
	os.remove(tmp_genome)
	"""mask regions """
	masked_sv = [(x[0],x[1],x[2],x[3],x[4]) for x in BedTool(annot_sv).subtract(BedTool('{}{}_excluded.bed.gz'.format(Config().resource_path(),gen)))]
	masked_start_flank = [(x[0],x[1],x[2],x[3],x[4]) for x in start_flank500bp.subtract(BedTool('{}{}_excluded.bed.gz'.format(Config().resource_path(),gen)))]
	masked_end_flank= [(x[0],x[1],x[2],x[3],x[4]) for x in end_flank500bp.subtract(BedTool('{}{}_excluded.bed.gz'.format(Config().resource_path(),gen)))]
	"""take the union of the two masked call sets"""
	if len(masked_sv) == 0:
		print 'FATAL ERROR: SVs completely overlap with excluded regions'
		sys.stderr.write('FATAL ERROR: SVs completely overlap with excluded regions\n')
		sys.exit(1)
	sv_tag = list(zip(*masked_sv)[4])
	passed_tag = list(set(sv_tag))
	masked_sv_dict=append_list2dict(masked_sv)
	masked_start_flank_dict=append_list2dict(masked_start_flank)
	masked_end_flank_dict=append_list2dict(masked_end_flank)
	master_sv={}
	for i in passed_tag:
		windows = (start_flank_dict[i],end_flank_dict[i])
		sv_span = masked_sv_dict[i]
		flank_span=None
		if masked_start_flank_dict.get(i) != None and masked_end_flank_dict.get(i) != None:
			flank_span = masked_start_flank_dict[i] + masked_end_flank_dict[i]
		master_sv[sv_dict[int(i)]]=(sv_span,flank_span,windows)
	return master_sv
cdef GC(nuc):
	cdef int A=0
	cdef int C=0
	cdef int G=0
	cdef int T=0
	cdef int N=0
	for x in nuc:
		A+=int(x[5])
		C+=int(x[6])
		G+=int(x[7])
		T+=int(x[8])
		N+=int(x[9])
	return ((G+C)*100)/(A+C+G+T+N)
cdef gc_sv(sv,master_sv,gen,Conf):
	sv_gc={}
	fasta = Conf.fasta(gen)
	for call in sv:
		if master_sv.get(call)==None: continue
		svs = master_sv[call][0]
		if Conf.fasta_chr_flag == False: svs = [tuple(map(lambda y: str.replace(str(y),'chr',''),x)) for x in svs]
		GC_content = int(5 * round(float(GC(BedTool(svs).nucleotide_content(fi=fasta)))/5))
		sv_gc[call]=GC_content
	return sv_gc
cdef is_discordant(Aln,windows,ci,matepos):
	"""returns True if Aln is discordant"""
	((s1,e1),(s2,e2)) = windows
	if abs(Aln.tlen) >= ci:
		"""insert size is greater than 5MAD from median"""
		if (int(s1) <= Aln.pos <= int(e1) and int(s2) <= int(matepos) <= int(e2)) or (int(s2) <= Aln.pos <= int(e2) and int(s1) <= int(matepos) <= int(e1)):
			return True
		else:
			return False
cdef is_split(Aln,windows,c,legacyM):
	"""returns True if Aln is split"""
	((s1,e1),(s2,e2)) = windows
	second_align=[]
	if legacyM == True:
		if Aln.is_secondary == True: second_align = Aln.get_tag("SA").split(',')
	else:
		if Aln.is_supplementary == True: second_align = Aln.get_tag("SA").split(',')
	if len(second_align) == 0: return False
	else:
		if second_align[0] != c: return False
		else:
			second_align[1] = int(second_align[1])
			if ( ((int(s1) <= Aln.pos <= int(e1)) and (int(s2) <= second_align[1]-1 <= int(e2)))
			 or ((int(s2) <= Aln.pos <= int(e2)) and (int(s1) <= second_align[1]-1 <= int(e1)))):
				"""supplementary/secondary alignment must be near the opposite breakpoint of the primary alignment"""
				return True
			else:
				return False
cdef normalize_coverage(double Aln_count, double span, double chr_cov, double Aln_length): return (((Aln_count/span)*Aln_length)/chr_cov)


#!/usr/bin/env python
from sv2_backend import check_PAR,errFH,format_chrom,tokenize 
from pybedtools import BedTool
import sys
__help_url__='https://github.com/dantaki/SV2/wiki/input#sv-input' 
class Structural_Variant():
	def __init__(self,sv=None,variant_id=None,gen=None):
		self.raw=[(str(x[0]),int(x[1]),int(x[2]),str(x[3])) for x in BedTool(list(set(sv))).sort()]
		self.variant_id=variant_id
		self.par={} # [locus]=True/False; true if intersects PAR 
		for locus in self.raw:
			chrom = format_chrom(locus[0])
			if chrom != 'chrX' and chrom != 'chrY': continue
			key = (chrom,int(locus[1]),int(locus[2]),str(locus[3]))
			self.par[key]=check_PAR('{} {} {}'.format(chrom,locus[1],locus[2]),gen) 
def checkCall(cl,l,fh):
	keep=True
	if 'DEL' not in cl and 'DUP' not in cl: 
		sys.stderr.write('WARNING: Skipping bad entry {} in {}. {} not an accepted SV type. Accepted types: DEL or DUP: {}\n'.format(l.rstrip(),fh,cl,__help_url__))
		keep=False
	return keep
def Bed(fh,sv):
	errFH(fh)
	with open(fh,'r') as f:
		for l in f:
			if l.startswith('#'): continue
			r = tokenize(l)
			if r==0 or len(r)<4: 
				sys.stderr.write('WARNING: Skipping bad entry {} in BED file {}. BED format requirements: {}\n'.format(l.rstrip(),fh,__help_url__))
			elif r!=0 and checkCall(str(r[3]),l,fh)!=False: sv.append((str(r[0]),int(r[1]),int(r[2]),str(r[3])))
def Vcf(fh,sv,variant_id):
	errFH(fh)
	vcf_fh=None
	if fh.endswith('.gz'):
		import gzip
		vcf_fh = gzip.open(fh,'r')
	else: vcf_fh = open(fh,'r')
	with vcf_fh as f:
		for l in f:
			if l.startswith('#'): continue
			s,e,r=-9,-9,tokenize(l)
			if r!=0:
				c,s=str(r[0]),int(r[1])
				s-=1
				if len(r)<8: sys.stderr.write('WARNING: Skipping bad entry {} in VCF file {}. INFO column is required: {}\n'.format(l.rstrip(),fh,__help_url__))
				else:
					for i in r[7].split(';'):
						if 'SVTYPE=' in i: cl=i.replace('SVTYPE=','')
						if i.startswith('END=')and 'CI' not in i: e=int(i.replace('END=',''))
					if s==-9 or e==-9: sys.stderr.write('WARNING: Skipping bad entry {} in VCF file {}. SV entry missing positions: {}\n'.format(l.rstrip(),fh,__help_url__))
					else:
						if checkCall(cl,l,fh)==True:
							sv.append((c,s,e,cl))
							variant_id[(str(c),int(s),int(e),str(cl))]=str(r[2])
	vcf_fh.close()
def skip_sv(c,s,e,cl,limit):
	skip=False
	if e<=s or s>=e: 
		skip=True
		sys.stderr.write('WARNING: end position is less than or equal to start position. Error found here: {}:{}-{} {}. Skipping ...\n'.format(c,s,e,cl))
	if s == limit or e >= limit:
		skip=True
		sys.stderr.write('WARNING: positions surpass reference limit. Error found here: {}:{}-{} {}. Skipping ...\n'.format(c,s,e,cl))
	if s<0 or e<0:
		skip=True
		sys.stderr.write('WARNING: positions are less than 0. Error found here: {}:{}-{} {}. Skipping ...\n'.format(c,s,e,cl))
	return skip
def sv_init(beds=None,vcfs=None,gen=None):
	sv,variant_id=[],{}
	if beds!=None:
		for fh in beds: Bed(fh,sv)
	if vcfs!=None:
		for fh in vcfs: Vcf(fh,sv,variant_id)
	if len(sv)<1: 
		print 'FATAL ERROR: SVs were not formatted correctly. See https://github.com/dantaki/SV2/wiki/input#sv-input for details'
		sys.stderr.write('FATAL ERROR: SVs were not formatted correctly. See https://github.com/dantaki/SV2/wiki/input#sv-input for details\n')
		sys.exit(1)
	SV = Structural_Variant(sv,variant_id,gen)
	return SV


#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
from sv2_backend import check_PAR,format_chrom
from Svm import biallelic_del_svm,biallelic_dup_svm,male_sex_chrom_svm
from Variant import Variant
import numpy as np
cimport numpy as np
import pandas as pd
import os,sys
np.import_array()
def init_dataframe(obj,header=['chr','start','end','type','size','id','old_covr','covr_GC','dpe','sr','SNP_coverage','HET_ratio','SNPs','HET_SNPs']):
	df = pd.DataFrame(obj)
	df.columns=header
	return df
def genotype(features,out,classifier_name,Ped,Conf,gen):
	SVs={} # SVs[SV.locus]=Variant object
	Conf.get_clf(classifier_name)
	biallelic_dels,biallelic_dups,male_sex_chrom_dels,male_sex_chrom_dups=partition_svs(features,Ped,gen)
	pd.options.mode.chained_assignment = None
	genos=[]
	if not out.endswith('.txt'): out = out+'.txt'
	ofh = open(out,'w')
	ofh.write('\t'.join(('#CHROM','START','END','TYPE','LENGTH','ID','COVERAGE','DISCORDANT_PAIRED-END','SPLIT_READS','SNV_COVERAGE','SNVs','HETEROZYGOUS_ALLELE_RATIO','HETEROZYGOUS_SNVs','COPY_NUMBER_GENOTYPE','REF_GENOTYPE_LIKELIHOOD','HET_GENOTYPE_LIKELIHOOD','HOM_GENOTYPE_LIKELIHOOD','ALT_GENOTYPE_LIKELIHOOD','REF_QUAL','ALT_QUAL','CLASSIFIER'))+'\n')
	if len(biallelic_dels)>0:
		c_biallele_dels(biallelic_dels,Ped,Conf,SVs,ofh)
	if len(biallelic_dups)>0:
		c_biallele_dups(biallelic_dups,Ped,Conf,SVs,ofh)
	if len(male_sex_chrom_dels)>0:
		c_male_sex_chrom_dels(male_sex_chrom_dels,Ped,Conf,SVs,ofh)
	if len(male_sex_chrom_dups)>0:
		c_male_sex_chrom_dups(male_sex_chrom_dups,Ped,Conf,SVs,ofh)
	ofh.close()
	for locus in SVs:
		Variant = SVs[locus]
		Variant.median_likelihood()
		Variant.add_standard_filters()
		Variant.add_denovo_filters()
	return SVs
def partition_svs(features,Ped,gen):
	biallelic_dels,biallelic_dups,male_sex_chrom_dels,male_sex_chrom_dups = [],[],[],[]
	for x in features:
		chrom,svtype,sample_id=format_chrom(x[0]),str(x[3]),str(x[5])
		if 'DEL' in svtype:
			if (chrom=='chrX' or chrom=='chrY') and Ped.males.get(sample_id)!=None and check_PAR('{} {} {}'.format(chrom,x[1],x[2]),gen)==False: male_sex_chrom_dels.append(x)
			else: biallelic_dels.append(x)
		elif 'DUP' in svtype:
			if (chrom=='chrX' or chrom=='chrY') and Ped.males.get(sample_id)!=None and check_PAR('{} {} {}'.format(chrom,x[1],x[2]),gen)==False: male_sex_chrom_dups.append(x)
			else: biallelic_dups.append(x)
	return biallelic_dels,biallelic_dups,male_sex_chrom_dels,male_sex_chrom_dups
cdef biallele(df,Ped,SVs,ofh):
	for x in df.values:
		ofh.write('\t'.join(map(str,x))+'\n')
		SV = Variant(x)
		if SVs.get(SV.locus)!=None: SV=SVs[SV.locus]
		clf=str(x[20])
		SV.clf.append(clf)
		cn = format(float(x[6])*2,'.2f')
		svgt = int(x[13])
		skip=False
		if svgt==2:
			lik=np.rint(float(x[18]))
			skip=set_biallelic_genotype(SV,x[0],x,Ped,'0/0',cn,lik)
			if skip==False:
				if SV.ref.get(clf)==None: SV.ref[clf]=[lik]
				else: SV.ref[clf].append(lik)
		else:
			lik=np.rint(float(x[19]))
			if svgt==1 or svgt==3: skip=set_biallelic_genotype(SV,x[0],x,Ped,'0/1',cn,lik)
			elif svgt==0 or svgt==4: skip=set_biallelic_genotype(SV,x[0],x,Ped,'1/1',cn,lik)
			if skip==False:
				
				if SV.alt.get(clf)==None: SV.alt[clf]=[lik]
				else: SV.alt[clf].append(lik)
		SVs[SV.locus]=SV
cdef c_biallele_dels(biallelic_dels,Ped,Conf,SVs,ofh):		
	biallele(biallelic_del_svm(init_dataframe(biallelic_dels),Conf.gtclf),Ped,SVs,ofh)	
cdef c_biallele_dups(biallelic_dups,Ped,Conf,SVs,ofh):
	biallele(biallelic_dup_svm(init_dataframe(biallelic_dups),Conf.gtclf),Ped,SVs,ofh)
cdef c_male_sex_chrom_dels(male_sex_chrom_dels,Ped,Conf,SVs,ofh):
	male_sex_chrom(male_sex_chrom_svm(init_dataframe(male_sex_chrom_dels),Conf.gtclf,'delmsc'),Ped,SVs,ofh)
cdef c_male_sex_chrom_dups(male_sex_chrom_dups,Ped,Conf,SVs,ofh):
	male_sex_chrom(male_sex_chrom_svm(init_dataframe(male_sex_chrom_dups),Conf.gtclf,'dupmsc'),Ped,SVs,ofh)
cdef male_sex_chrom(df,Ped,SVs,ofh):
	for x in df.values:
		ofh.write('\t'.join(map(str,x))+'\n')
		SV = Variant(x)
		if SVs.get(SV.locus)!=None: SV=SVs[SV.locus]
		clf=str(x[20])
		SV.clf.append(clf)
		cn = format(float(x[6]),'.2f')
		svgt = int(x[13])
		if svgt==1:
			lik=np.rint(float(x[18]))
			SV.genotype(x,'0',cn,lik,False)
			if SV.ref.get(clf)==None: SV.ref[clf]=[lik]
			else: SV.ref[clf].append(lik)
		else:
			lik=np.rint(float(x[19]))
			SV.genotype(x,'1',cn,lik,False)
			if SV.alt.get(clf)==None: SV.alt[clf]=[lik]
			else: SV.alt[clf].append(lik)
		SVs[SV.locus]=SV
cdef set_biallelic_genotype(SV,chrom,data,Ped,genotype,copy_number,lik):
	skip=False
	SV.genotype(data,genotype,copy_number,lik)
	sample_id=str(data[5])
	if Ped.males.get(sample_id)==None and format_chrom(chrom) == 'chrY':
		skip=True
		SV.genotype(data,'.',copy_number,lik)
	return skip


#!/usr/bin/env python2
from datetime import timedelta
from pybedtools import BedTool
from time import time
import json,os,random,string,sys
def accepted_chrom(flag,gen):
	chroms,prefix=[],''
	if flag==True: prefix='chr'
	if 'hg' in gen:
		for i in range(22): chroms.append(prefix+str(i+1))
	else:
		for i in range(19): chroms.append(prefix+str(i+1))
	chroms.append(prefix+'X')
	chroms.append(prefix+'Y')
	return chroms
def check_ids(ped,feats):
	missing_feats, missing_ped = list(set(ped)-set(feats)),list(set(feats)-set(ped))
	if len(missing_feats)>0:
		for x in missing_feats: sys.stderr.write('WARNING: sample {} in ped file is missing features. Skipping {}\n'.format(x,x))
	if len(missing_ped)>0:
		for x in missing_ped: sys.stderr.write('WARNING: sample {} has features but is missing in the ped file. Skipping {}\n'.format(x,x))
	return sorted(list(set(feats).intersection(ped)),key=str.lower)
def check_PAR(sv,gen):
	from sv2Config import Config
	in_par = False
	if len(BedTool(sv,from_string=True).intersect(BedTool('{}par_{}.bed'.format(Config().resource_path(),gen)),f=0.5)) > 0: in_par=True
	return in_par
def dump_json(o,d):
	ofh=open(o,'w')
	json.dump(d,ofh,indent=4)
	ofh.close()
def errFH(fh):
	if not os.path.isfile(fh):
		print 'FATAL ERROR {} does not exist'.format(fh)
		sys.stderr.write('FATAL ERROR {} does not exist\n'.format(fh))
		sys.exit(1)
def format_chrom(c):
	if 'chr' not in c: c='chr'+str(c)
	c = c.replace('chr23','chrX').replace('chr24','chrY')
	return c
def get_path():
	try:
		root = __file__
		if os.path.islink(root): root = os.path.realpath(root)
		return os.path.dirname(os.path.abspath(root))
	except:
		print 'FATAL ERROR {} NOT FOUND\n'.format(__file__)
		sys.stderr.write('FATAL ERROR {} NOT FOUND\n'.format(__file__))
		sys.exit(1)
def make_dir(mydir):
	if not os.path.exists(mydir): os.makedirs(mydir)
def mask_bed(fh,gen):
	from sv2Config import Config
	return BedTool(fh).subtract('{}{}_excluded.bed.gz'.format(Config().resource_path(),gen))
def match_chrom_key(call,Bam,gen):
	chrom_key = format_chrom(str(call[0]))
	if Bam.sex == 1 and (chrom_key=='chrX' or chrom_key=='chrY') and check_PAR('{} {} {}'.format(chrom_key,call[1],call[2]),gen)== True: chrom_key='GENOME'
	return chrom_key
def match_chrom_prefix(c,flag):
	if flag==False and c.startswith('chr'): c=c.replace('chr','')
	elif flag==True and not c.startswith('chr'): c='chr'+c
	return c
def query_yes_no(question, default="yes"):
	"""https://code.activestate.com/recipes/577058/"""
	valid = {"yes": True, "y": True, "ye": True,"no": False, "n": False}
	if default is None: prompt = " [y/n] "
	elif default == "yes": prompt = " [Y/n] "
	elif default == "no": prompt = " [y/N] "
	else: raise ValueError("invalid default answer: '%s'" % default)
	while True:
		sys.stdout.write(question + prompt)
		choice = raw_input().lower()
		if default is not None and choice == '': return valid[default]
		elif choice in valid: return valid[choice]
		else: sys.stdout.write("Please respond with 'yes' or 'no' (or 'y' or 'n')\n")
def rand_id(size=7, chars=string.ascii_uppercase + string.digits):
	return ''.join(random.choice(chars) for _ in range(size))
def reciprocal_overlap(x):
	(s1,s2,e1,e2) = x
	sz1 = e1-s1
	sz2 = e2-s2
	smax=max(s1,s2)
	emin=min(e1,e2)
	ovr = emin-smax
	return min((ovr/sz1),(ovr/sz2))
def report_time(init_time,s):
	f=[]
	s = ' '+s+' <time elapsed: {}>'.format(timedelta(seconds=int(time())-init_time))
	for _ in range(len(s)+1): f.append('*')
	print ''.join(f)+'\n'+s+'\n'+''.join(f)
def slash_check(mydir):
	if not mydir.endswith('/'): mydir = mydir+'/'
	if not os.path.isdir(mydir):
		print 'FATAL ERROR: {} not found.\n'.format(mydir)
		sys.stderr.write('FATAL ERROR: {} not found.\n'.format(mydir))
		sys.exit(1)
	return mydir
def tokenize(l):
	r=l.rstrip().split('\t')
	if len(r)==1: r=l.rstrip().split(' ')
	if len(r)==1:
		sys.stderr.write('WARNING: {} not tokenized correctly. Check if {} is Tab or Space delimited\n'.format(l,l))
		return 0
	else: return r


#!/usr/bin/env python2
from sv2_backend import format_chrom
from Genotype import init_dataframe,partition_svs
from Svm import training_features
from Vcf import VCF
import datetime
def output(Structural_Variant,SVs,Ped,ids,gen,ofh,anno_flag,tmp_dir):
	Ofh = VCF(ofh)
	Ofh.init_header(datetime.date.today(),ids,Structural_Variant,gen)
	Ofh.load_genotypes(Structural_Variant,SVs,Ped,ids,gen,anno_flag,tmp_dir)
	if not ofh.endswith('.vcf'): ofh = ofh+'.vcf'
	vcf_ofh = open(ofh,'w')
	vcf_ofh.write('\n'.join(Ofh.head)+'\n')
	chroms={}
	for entry in Structural_Variant.raw:
		chroms[entry[0]]=1
		variant_id = '.'
		if Structural_Variant.variant_id.get(entry)!=None: variant_id=Structural_Variant.variant_id[entry]
		locus=(format_chrom(entry[0]),int(entry[1]),int(entry[2]),str(entry[3]))
		genotypes=[]
		if Ofh.genotypes.get(locus)==None:
			for sample_id in ids:
				gt='./.'
				if (locus[0]=='chrX' or locus[0]=='chrY') and Ped.males.get(sample_id)!=None and Structural_Variant.par[locus]==False: gt='.'
				genotypes.append(gt)
		else: genotypes=Ofh.genotypes[locus]
		genotypes = '\t'.join(genotypes)
		out=Ofh.init_row(entry)
		vcf_ofh.write('{}\t{}\t{}\t.\t{}\t{}\n'.format(entry[0],int(entry[1])+1,variant_id,out,genotypes))
	vcf_ofh.close()
def sv2_train_output(feats,Ped,gen,opre):
	biallelic_dels,biallelic_dups,male_sex_chrom_dels,male_sex_chrom_dups = partition_svs(feats,Ped,gen)
	if len(biallelic_dels)>0: training_features(init_dataframe(biallelic_dels),0,opre)
	if len(biallelic_dups)>0: training_features(init_dataframe(biallelic_dups),1,opre)
	if len(male_sex_chrom_dels)>0: training_features(init_dataframe(male_sex_chrom_dels),2,opre)
	if len(male_sex_chrom_dups)>0: training_features(init_dataframe(male_sex_chrom_dups),3,opre)


#!/usr/bin/env python
from sv2_backend import accepted_chrom,errFH,format_chrom,make_dir
from collections import OrderedDict
import pysam,sys
class Bam():
	def __init__(self,fh=None,Ped=None,gen=None):
		self.id = None
		self.chr_flag=False # True == 'chrN'; False == 'N'
		self.sex=None
		self.char='b'
		self.refs=OrderedDict()
		self.fh=fh
		self.Snv=None
		self.snv_index=None
		sm=[]
		bam = pysam.AlignmentFile(fh)
		if bam.is_cram==True: self.char='c'
		header = bam.header
		if header.get('RG')==None: sys.stderr.write('WARNING: {} lacks Read Group (@RG) entry in the header. Skipping ...\n'.format(fh))
		else:
			for entry in header['RG']:
				if entry.get('SM')!=None: sm.append(entry['SM'])
			sm = list(set(sm))
			if len(sm)>1: sys.stderr.write('WARNING: {} contains two sample entries ({}) in the header. Skipping {} ...\n'.format(fh,sm,fh))
			else: self.id=sm[0]
		if self.id!=None:
			if Ped.sex.get(self.id)==None: sys.stderr.write('WARNING: {} is not found in the PED file. Skipping {} ...\n'.format(self.id,fh))
			else:
				self.sex=Ped.sex[self.id]
				if str(bam.references[0]).startswith('chr'): self.chr_flag=True
				chroms = accepted_chrom(self.chr_flag,gen)
				for i in range(len(bam.references)):
					chrom,leng = str(bam.references[i]),bam.lengths[i]
					if chrom in chroms: self.refs[chrom]=int(leng)
		bam.close()
	def tmp_chrom_file(self,tmp_dir=None,genome=True,chrom=None):
		make_dir(tmp_dir)
		tmp_genome = tmp_dir+'{}.genome'.format(self.id)
		tmpfh = open(tmp_genome,'w')
		if genome==True:
			if chrom==None:
				for ref in self.refs: tmpfh.write('{}\t{}\n'.format(format_chrom(ref),self.refs[ref]))
			else: tmpfh.write('{}\t{}\n'.format(format_chrom(chrom),self.refs[chrom]))
		if genome==False:
			for ref in self.refs:tmpfh.write('{}\t0\t{}\n'.format(format_chrom(ref),self.refs[ref]))
		tmpfh.close()
		return tmp_genome
def bam_init(args=None,Ped=None,Snv=None,gen=None):
	bams=[]
	for f in args:
		errFH(f)
		try: pysam.AlignmentFile(f).check_index()
		except ValueError:
			sys.stderr.write('WARNING: {} is not indexed with samtools index. Skipping ...\n'.format(f))
			continue
		except AttributeError:
			sys.stderr.write('WARNING: {} appears to be in SAM format. Convert to BAM with `samtools view -bh {}` and index with samtools index\n'.format(f,f))
			continue
		bam = Bam(f,Ped,gen)
		if len(Snv)<1:
			print 'FATAL ERROR: SNV file(s) were not formatted correctly. See https://github.com/dantaki/SV2/wiki/input#snv-vcf for details'
			sys.stderr.write('FATAL ERROR: SNV file(s) were not formatted correctly. See https://github.com/dantaki/SV2/wiki/input#snv-vcf for details\n')
			sys.exit(1)
		for snv in Snv:
			if snv.id.get(bam.id)!=None:
				bam.Snv,bam.snv_index=snv,snv.id[bam.id]
		if bam.id!=None and bam.Snv==None: sys.stderr.write('WARNING: BAM file {} sample name (@RG  SM:<sample_id>):{} not found in SNV VCFs. Skipping {} ...\n'.format(f,bam.id,f))
		if bam.id==None: sys.stderr.write('WARNING: Skipping BAM file {}. No sample name (@RG SM:<sample_id>). See https://github.com/dantaki/SV2/wiki/input#bam for details')
		if bam.id!=None and bam.Snv!=None: bams.append(bam)
	if len(bams)<1:
		print 'FATAL ERROR: BAM file(s) were not formatted correctly. See https://github.com/dantaki/SV2/wiki/input#bam for details'
		sys.stderr.write('FATAL ERROR: BAM file(s) were not formatted correctly. See https://github.com/dantaki/SV2/wiki/input#bam for details\n')
		sys.exit(1)
	return bams


#!/usr/bin/env python2
'''
Copyright <2017> <Danny Antaki>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''
__version__='1.4.3'
from sv2_backend import check_ids,get_path,make_dir,rand_id,report_time,slash_check
from Bam import bam_init
from sv2Config import Config
from FeatureExtraction import extract_feats
from Genotype import genotype
from Merge import merge_sv
from Output import output
from Ped import ped_init
from Preprocess import preprocess
from Snv import snv_init
from SV import sv_init
from argparse import RawTextHelpFormatter
from glob import glob
from time import time
import shutil,sys,os,argparse
splash='                       ____\n  _____________   ___ |___ \\\n /   _____/\   \ /   // ___/\n \_____  \  \   Y   //_____)\n /        \  \     /\n/_________/   \___/\n'
__useage___="""
Support Vector Structural Variation Genotyper
Version {}    Author: Danny Antaki <dantaki at ucsd dot edu>    github.com/dantaki/SV2

  sv2 -i <bam ...> -v <vcf ...> -b <bed ...> -snv <snv vcf ...> -p <ped ...> [OPTIONS]

input arguments: github.com/dantaki/SV2/wiki/Options#input-arguments

  Note: input arguments can take multiple files, separated by space <-i BAM1 BAM2 ...>

  -i, -bam    ...     bam file(s)
  -v, -vcf    ...     vcf files(s) of SVs
  -b, -bed    ...     bed files(s) of SVs
  -snv        ...     snv vcf files(s), must be bgzipped and tabixed
  -p, -ped    ...     ped files(s)

genotype arguments: github.com/dantaki/SV2/wiki/Options#genotype-arguments

  -g, -genome         reference genome build <hg19, hg38, mm10> [default: hg19]
  -pcrfree            GC content normalization for pcr free sequences
  -M                  bwa mem -M compatibility, split-reads flagged as secondary instead of supplementary
  -merge              merge overlapping SV breakpoints after genotyping
  -min-ovr            minimum reciprocal overlap for merging [default: 0.8]
  -no-anno            genotype without annotating variants   

  -pre                preprocessing output directory, skips preprocessing
  -feats              feature extraction output directory, skips feature extraction

classifier arguments: github.com/dantaki/SV2/wiki/Options#classifier-arguments

  -load-clf           add custom classifiers (-load-clf <clf.json>)
  -clf                define classifier for genotyping [default:default]

config arguments: github.com/dantaki/SV2/wiki/Options#config-arguments

  -download           download required data files
  -hg19               hg19 fasta file
  -hg38               hg38 fasta file
  -mm10               mm10 fasta file
  
optional arguments:
 
  -L, -log            log file for standard error messages [default: STDOUT]
  -T, -tmp-dir        directory for temporary files [default: working directory]
  -s, -seed           random seed for preprocessing shuffling [default: 42]
  -o, -out            output prefix [default: sv2_genotypes]
  -O, -odir           output path, location for sv2 output directories [default: working directory]

  -h, -help           show this message and exit
""".format(__version__)
def main():
	init_time = int(time())
 	parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter,usage=splash.replace("       ","",1)+__useage___,add_help=False)
	inArgs, genoArgs,clfArgs,configArgs,optArgs= parser.add_argument_group('input arguments'),parser.add_argument_group('genotype arguments'), parser.add_argument_group('classifier arguments'),parser.add_argument_group('config arguments'),parser.add_argument_group('optional arguments')
	inArgs.add_argument('-i','-bam',type=str,default=None,nargs='*')
	inArgs.add_argument('-b','-bed',type=str,default=None,nargs='*')
	inArgs.add_argument('-v','-vcf',type=str,default=None,nargs='*')
	inArgs.add_argument('-snv',type=str,default=None,nargs='*')
	inArgs.add_argument('-p','-ped',type=str,default=None,nargs='*')
	genoArgs.add_argument('-g','-genome',required=False,default='hg19',type=str)
	genoArgs.add_argument('-pcrfree',required=False,default=False,action="store_true")
	genoArgs.add_argument('-M',default=False,required=False,action="store_true")
	genoArgs.add_argument('-merge',required=False,default=False,action="store_true")
	genoArgs.add_argument('-min-ovr',required=False,default=None,type=float)
	genoArgs.add_argument('-no-anno',required=False,default=False,action="store_true")
	genoArgs.add_argument('-pre',required=False,default=None)
	genoArgs.add_argument('-feats',required=False,default=None)
	clfArgs.add_argument('-load-clf',required=False, default=None,type=str)
	clfArgs.add_argument('-clf',required=False,default='default',type=str)
	configArgs.add_argument('-download',default=False,required=False,action="store_true")
	configArgs.add_argument('-hg19',default=None,required=False)
	configArgs.add_argument('-hg38',default=None,required=False)	
	configArgs.add_argument('-mm10',default=None,required=False)
	optArgs.add_argument('-L','-log',default=None,required=False)
	optArgs.add_argument('-T','-tmp-dir',default=os.getcwd()+'/sv2_tmp_'+rand_id(),required=False)
	optArgs.add_argument('-s','-seed',required=False,default=42,type=int)
	optArgs.add_argument('-o','-out',required=False,default="sv2_genotypes.vcf",type=str)
	optArgs.add_argument('-O','-odir',required=False,default=os.getcwd(),type=str)
	optArgs.add_argument('-h','-help',required=False,action="store_true",default=False)
	args = parser.parse_args()
	bams,bed,vcf,snv,ped = args.i,args.b,args.v,args.snv,args.p
	gen,pcrfree,legacy_m,merge_flag,min_ovr,anno_flag= args.g,args.pcrfree,args.M,args.merge,args.min_ovr,args.no_anno
	predir,featsdir = args.pre,args.feats
	clfLoad, classifier_name = args.load_clf,args.clf
	download, conf_hg19,conf_hg38, conf_mm10=args.download,args.hg19,args.hg38,args.mm10
	logfh, tmp_dir, seed, ofh, odir = args.L,args.T,args.s,args.o,args.O
	_help = args.h
	if (_help==True or len(sys.argv)==1):
		print splash+__useage___
		sys.exit(0)
	if logfh!=None:
		lfh = open(logfh,'w')
		sys.stderr=lfh
	preprocess_files={}
	feats_files={}
	gens = ['hg19','hg38','mm10']
	olog = logfh
	if olog == None: olog = 'STDOUT'
	print 'sv2 version:{}    report bugs to <dantaki at ucsd dot edu>       error messages located in {}'.format(__version__,olog)
	Confs=Config()
	if clfLoad!=None:
		Confs.load_clf(clfLoad)
		sys.exit(0)
	if download==True:
		Confs.download_resource()
		sys.exit(0)
	if conf_hg19 != None or conf_hg38 != None or conf_mm10 != None:
		Confs.write_config(conf_hg19,conf_hg38,conf_mm10)
		sys.exit(0)
	if bams==None and predir==None and featsdir==None:
		print 'FATAL ERROR: No BAM file specified <-i, -bam  FILE ...>'
		sys.stderr.write('FATAL ERROR: No BAM file specified <-i, -bam  FILE ...>\n')
		sys.exit(1)
	if snv==None and predir==None and featsdir==None:
		print 'FATAL ERROR: No SNV VCF file specified <-snv  FILE ...>'
		sys.stderr.write('FATAL ERROR: No SNV VCF file specified <-snv  FILE ...>\n')
		sys.exit(1)
	if ped==None:
		print 'FATAL ERROR: No PED file specified <-p, -ped  FILE ...>'
		sys.stderr.write('FATAL ERROR: No PED file specified <-p, -ped  FILE ...>\n')
		sys.exit(1)
	if bed==None and vcf==None:
		print 'FATAL ERROR: No SVs provided <-b, -bed  BED ...> <-v,-vcf  VCF ...>'
		sys.stderr.write('FATAL ERROR: No SVs provided <-b, -bed  BED ...> <-v,-vcf  VCF ...>\n')
		sys.exit(1)
	if gen not in gens:
		print 'FATAL ERROR -g must be hg19 or hg38. NOT {}'.format(gen)
		sys.stderr.write('FATAL ERROR -g must be hg19 or hg38. NOT {}\n'.format(gen))
		sys.exit(1)
	Confs.check_config()
	Peds=ped_init(ped)
	if bams!=None: Bams=bam_init(bams,Peds,snv_init(snv),gen)
	SV = sv_init(bed,vcf,gen)
	ofh = ofh.replace('.txt','.vcf').replace('.out','.vcf')
	if not ofh.endswith('.vcf'): ofh=ofh+'.vcf'
	make_dir(tmp_dir)
	tmp_dir=slash_check(tmp_dir)
	if not odir.endswith('/'): odir = odir+'/'
	make_dir(odir)
	"""
	PREPROCESSING
	"""
	if predir == None  and featsdir==None:
		outdir = odir+'sv2_preprocessing/'
		make_dir(outdir)
		for bam in Bams:
			preofh = outdir+bam.id+'_sv2_preprocessing.txt'
			preprocess_files[bam.id]=preofh
			preprocess(bam,preofh,seed,gen,tmp_dir)
	elif predir!=None:
		predir=slash_check(predir)
		for fh in glob(predir+'*sv2_preprocessing.txt'):
			f = open(fh)
			if sum(1 for l in open(fh)) <= 1: continue
			else:
				preids=[]
				for l in f:
					if l.startswith('#'):continue
					preids.append(l.rstrip('\n').split('\t').pop(0))
			f.close()
			for iid in set(preids):
				if iid in Peds.ids : preprocess_files[iid]=fh
	report_time(init_time,'PREPROCESSING COMPLETE')
	""""
	FEATURE EXTRACTION
	"""
	if featsdir == None:
		outdir = odir+'sv2_features/'
		make_dir(outdir)
		for bam in Bams:
			if preprocess_files.get(bam.id) == None:
				sys.stderr.write('WARNING: BAM sample id {} not found in preprocessing files. Skipping ...\n'.format(bam.id))
				continue
			prefh = preprocess_files[bam.id]
			featfh = outdir+bam.id+'_sv2_features.txt'
			feats_files[bam.id]=featfh
			extract_feats(bam,SV.raw,prefh,featfh,gen,pcrfree,legacy_m,Confs,tmp_dir)
	else:
		featsdir=slash_check(featsdir)
		for fh in glob(featsdir+'*sv2_features.txt'):
			f = open(fh)
			if sum(1 for l in open(fh)) <= 1: continue
			else:
				featsid=[]
				for l in f:
					if l.startswith('#'):continue
					featsid.append(l.rstrip('\n').split('\t').pop(5))
			f.close()
			for iid in set(featsid):
				if iid in Peds.ids : feats_files[iid]=fh
	report_time(init_time,'FEATURE EXTRACTION COMPLETE')	
	"""
	GENOTYPING
	"""
	outdir = odir+'sv2_genotypes/'
	make_dir(outdir)
	if merge_flag==True and min_ovr==None: min_ovr=0.8
	if min_ovr!=None: merge_flag=True
	feats,ids=[],[]
	for iid in feats_files:
		ids.append(iid)
		with open(feats_files[iid]) as f:
			for l in f:
				if l.startswith('#'):continue
				feats.append(tuple(l.rstrip('\n').split('\t')))
	ids = check_ids(Peds.ids,ids)
	SVs = genotype(feats,outdir+ofh.replace('.vcf','.txt'),classifier_name,Peds,Confs,gen)
	if merge_flag==True: SV.raw = merge_sv(SV.raw,SVs,min_ovr)
	output(SV,SVs,Peds,ids,gen,outdir+ofh,anno_flag,tmp_dir)
	shutil.rmtree(tmp_dir)
	if logfh!=None: lfh.close()
	report_time(init_time,'GENOTYPING COMPLETE')


#!/usr/bin/env python
from sv2_backend import format_chrom
from pybedtools import BedTool
import itertools
def check_merge(sv,minOvr):
	to_merge=[]
	for x in sv.intersect(sv,f=minOvr,F=minOvr,wao=True):
		x=tuple(x)
		if skip_sv(x)==False:  to_merge.append(x)
	return  to_merge
def first_merge(sv,minOvr):
	to_merge=[]
	for x in sv.intersect(sv,f=minOvr,F=minOvr,wao=True):
		x=tuple(x)
		k1,k2 = tokenize_sv(x),tokenize_sv(x,False)
		if k1 != k2:
			 to_merge.append(k1)
			 to_merge.append(k2)
	return [x for x in sv if tokenize_sv(tuple(x)) not in  to_merge]
def merging(sv,non):
	merged,trash=[],[]
	for x in sv:
		k1,k2 = tokenize_sv(x),tokenize_sv(x,False)
		non1,non2=0,0
		if non.get(k1)!=None: non1=non[k1]
		if non.get(k2)!=None: non2=non[k2]
		if non1=='NA':non1=0
		if non2=='NA':non2=0
		if non1 > non2:
			if k1 not in trash: merged.append(k1)
			trash.append(k2)
		elif non2 > non1:
			if k2 not in trash: merged.append(k2)
			trash.append(k1)
		else:
			if k1 not in trash: merged.append(k1)
			trash.append(k2)
	return BedTool(list(set(merged)))
def merge_sv(raw,SVs,minOvr):
	non,result,merged={},{},[]
	for locus in SVs: non[locus]=SVs[locus].med_alt
	dels = BedTool([(format_chrom(x[0]),x[1],x[2],x[3]) for x in raw if 'DEL' in x[3]])
	dups = BedTool([(format_chrom(x[0]),x[1],x[2],x[3]) for x in raw if 'DUP' in x[3]])
	dels_nomerge = first_merge(dels,minOvr)
	while len(check_merge(dels,minOvr))>0:
		dels = merging(check_merge(dels,minOvr),non)
	dups_nomerge = first_merge(dups,minOvr)
	while len(check_merge(dups,minOvr))>0:
		dups = merging(check_merge(dups,minOvr),non)
	for sv in list(itertools.chain(dels,dels_nomerge,dups,dups_nomerge)): 
		sv=tokenize_sv(tuple(sv))
		result[sv]=1
	for sv in raw:
		chrom,start,end,svtype = sv
                _sv = (format_chrom(chrom),start,end,svtype)
                if result.get(_sv)!=None: merged.append(sv) 
	return merged
def skip_sv(x):
	k1,k2 = tokenize_sv(x),tokenize_sv(x,False)
	if k1==k2: return True
	else: return False
def tokenize_sv(x,first_entry=True):
	if first_entry==True: return (str(x[0]),int(x[1]),int(x[2]),str(x[3]))
	else: return (str(x[4]),int(x[5]),int(x[6]),str(x[7]))



#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
from sv2_backend import errFH,format_chrom,match_chrom_key,match_chrom_prefix
import numpy as np
import os,pysam,sys
from pybedtools import BedTool
cimport numpy as np
class Snv():
	def __init__(self,fh=None):
		isHC,isFB=False,False
		self.fh=fh
		self.id={} # id[IID]=ROW INDEX
		self.chr_flag=False # True == 'chrN'; False == 'N'
		self.ad_format=None
		tbx = pysam.TabixFile(fh)
		header = list(tbx.header)
		if [x for x in header if 'FORMAT=<ID=AD' in x]: isHC=True
		if [x for x in header if 'FORMAT=<ID=DPR' in x]: isFB=True
		if isHC==isFB: sys.stderr.write('WARNING:{} SNV VCF does not contain Allele Depth. See details for SNV VCF format: github.com/dantaki/SV2/wiki/input#snv-vcf\nSkipping {} ...'.format(fh,fh))
		else:
			if isHC==True:self.ad_format='AD'
			if isFB==True:self.ad_format='DPR'
			head = [x for x in header if x.startswith('#CHROM')]
			if len(head)<1:
				sys.stderr.write('WARNING:{} SNV VCF does not contain a proper header. See details for SNV VCF format: github.com/dantaki/SV2/wiki/input#snv-vcf\nSkipping {} ...'.format(fh,fh))
				self.ad_format=None
			else:
				head = head[0].split('\t')
				for index in range(9,len(head)): self.id[head[index]]=index
				if str(tbx.contigs[0]).startswith('chr'): self.chr_flag=True
		tbx.close()		
def extract_snv_features(Bam,svs,master_sv,Pre,gen): return c_extract_snv_features(Bam,svs,master_sv,Pre,gen)
def preprocess_snv(Bam,masked): return c_preprocess_snv(Bam,masked)
def snv_init(l=None):
	snv=[]
	for fh in l:
		errFH(fh)
		try: pysam.TabixFile(fh)
		except IOError:
			sys.stderr.write('WARNING: tabix index file for {} not found. SNV files must be bgzipped and indexed with tabix. Skipping ...\n'.format(fh))
			continue
		Vcf = Snv(fh)
		if Vcf.ad_format!=None: snv.append(Vcf)
	return snv
cdef c_extract_snv_features(Bam,svs,master_sv,Pre,gen):
	cdef unsigned int s=0
	cdef unsigned int e=0
	cdef unsigned int depth=0
	cdef double ratio=0.
	cdef double norm_cov=0.
	snv_feats,het_feats={},{}
	Itr = pysam.TabixFile(Bam.Snv.fh)
	for call in svs:
		# foreach SV
		if master_sv.get(call)==None: continue
		sv_span,flank_span,windows = master_sv[call]
		locus_depth, locus_had=[],[]
		chrom = format_chrom(call[0])
		chrom_key = match_chrom_key(call,Bam,gen)
		if Pre.snv_cov[(Bam.id,chrom_key)]==None: continue
		else: norm_cov=Pre.snv_cov[(Bam.id,chrom_key)]
		#####################
		for locus in sv_span:
			# foreach masked region within SV
			c,s,e = str(locus[0]),int(locus[1]),int(locus[2])
			c=match_chrom_prefix(c,Bam.Snv.chr_flag)
			for Variant in Itr.fetch(region='{}:{}-{}'.format(c,s+1,e)):
				data= tokenize_vcf(Bam,Variant)
				if data==0:continue
				depth,ratio= data
				locus_depth.append(depth)
				if ratio!=-9:locus_had.append(ratio)
		#####################
		if len(locus_depth)==0: snv_feats[call]=(float('nan'),0)
		else: snv_feats[call]=(np.nanmedian(locus_depth)/norm_cov, len(locus_depth)) # normalized coverage, # snvs
		if len(locus_had)==0: het_feats[call]=(float('nan'),0)
		else: het_feats[call]=(np.nanmedian(locus_had),len(locus_had)) # median heterozygous allele depth, # heterozygous snvs
	Itr.close()
	return snv_feats,het_feats
cdef c_preprocess_snv(Bam,masked):
	cdef int genome_lens=0
	genome_depths=[]
	depths,snv_depth,genome={},{},[] # depth[ref]=[], snv_depth[ref]= median depth, # snvs
	Itr = pysam.TabixFile(Bam.Snv.fh)
	for entry in masked:
		c = str(entry[0])
		if Bam.chr_flag==False: c=c.replace('chr','')
		chrom = format_chrom(c)
		if depths.get(chrom)==None: depths[chrom]=[]
		region = '{}:{}-{}'.format(c,int(entry[1])+1,int(entry[2]))
		try:
			for Variant in Itr.fetch(region=region):
				data = tokenize_vcf(Bam,Variant,False) # False=only process depth; output is (depth, allele depth)
				if data == 0: continue
				depths[chrom].append(data[0])
		except ValueError: sys.stderr.write('WARNING: skipping {}. Region not found in {}\n'.format(region,Bam.Snv.fh))
	Itr.close()
	for chrom in depths:
		if len(depths[chrom])>0:
			snv_depth[chrom]= np.nanmedian(depths[chrom]),len(depths[chrom])
			genome_depths+=depths[chrom]
			genome_lens+=len(depths[chrom])
		else: snv_depth[chrom]=0,0
	if len(genome_depths)>0: snv_depth['GENOME']=np.nanmedian(genome_depths),genome_lens
	else: snv_depth['GENOME']=0,0
	return snv_depth	
cdef format_genotype(x): return x.replace('|','/')
cdef tokenize_vcf(Bam,Variant,HAD=True):
	cdef double ratio=-9
	cdef unsigned int depth
	cdef short depth_index=-9
	cdef short genotype_index=-9
	cdef short allele_depth_index=-9
	record=Variant.split('\t')
	try: record[Bam.snv_index]
	except IndexError:
		sys.stderr.write('WARNING: {} SNV VCF entry is malformed. Skipping ...\n'.format('\t'.join(record)))
		return 0
	if len(record) < Bam.snv_index: return 0
	c,s  = format_chrom(record[0]),int(record[1])
	_format = record[8].split(':')
	if HAD == True: # process allele depth and depth
		for ind in xrange(len(_format)):
			if _format[ind]=='DP': depth_index=ind
			if _format[ind]=='GT': genotype_index=ind
			if _format[ind]== Bam.Snv.ad_format: allele_depth_index=ind
		if depth_index==-9 or genotype_index==-9 or allele_depth_index == -9: return 0
		sample_entry = record[Bam.snv_index].split(':')
		if '.' in sample_entry[genotype_index] or '.' in sample_entry[depth_index]: return 0
		genotype=[int(x) for x in format_genotype(sample_entry[genotype_index]).split('/')]
		if len(genotype)!=2: return 0
		depth = int(sample_entry[depth_index])
		if genotype[0]!=genotype[1]:
			allele_depth = [float(x) for x in sample_entry[allele_depth_index].split(',')]
			a,b = allele_depth[genotype[0]],allele_depth[genotype[1]]
			if a== 0.0 or b == 0.0: return 0
			ratio = b/a
			if b > a: ratio=a/b
	else: # only process depth
		for ind in xrange(len(_format)):
			if _format[ind]=='DP': depth_index=ind
			if _format[ind]=='GT': genotype_index=ind
		if depth_index==-9: return 0
		sample_entry = record[Bam.snv_index].split(':')
		if '.' in sample_entry[genotype_index] or '.' in sample_entry[depth_index]:return 0
		depth=int(sample_entry[depth_index])
	return (depth,ratio)


#!/usr/bin/env python
from sv2_backend import format_chrom
import numpy as np
class Variant:
	def __init__(self,data=None):
		self.locus=(format_chrom(data[0]),int(data[1]),int(data[2]),str(data[3])) # (chrom,start,end,type)
		self.ref={} # [classifier_name]=[reference genotype likelihoods]
		self.med_ref='NA'
		self.alt={} # [classifier_name]=[alternate genotype likelihoods]
		self.med_alt='NA'
		self.format={} # [locus+id]='GT:CN:PE:SR:SC:SP:AR:HT:SQ:GL'
		self.gt={} # gt[locus+id]=genotype
		self.gq={} # gq[locus+id]=(ref,het,hom) likelihoods
		self.breakpoint_feats=0 # breakpoint features: discordant paired-ends + split-reads
		self.clf=[] # classifiers
		self.standard_filter='PASS'
		self.denovo_filter='PASS'
	def add_denovo_filters(self):
		yea,nay = 0,0
		for clfname in self.alt:
			non,feat,svlen =  self.alt[clfname],self.breakpoint_feats,self.locus[2]-self.locus[1]
			passflag=0
			ref=20
			if self.ref.get(clfname)!=None: ref=self.ref[clfname]
			if 'DEL' in clfname and 'male_sex_chrom' not in clfname:
				if feat != 0 and non >=12 and ref >=12: passflag=1
				elif feat==0 and 1000<=svlen<3000 and non >=24 and ref >=20: passflag=1
				elif feat==0 and 3000<=svlen<5000 and non >=20 and ref >=20: passflag=1
				elif feat==0 and svlen>= 5000 and non >=20 and ref >=18: passflag=1	
			elif 'DUP_breakpoint' in clfname:
				if feat != 0 and non >=11 and ref >=11: passflag=1
				if feat==0 and 3000<=svlen<100000 and non >=10 and ref >=15: passflag=1
				if feat==0 and  svlen>=100000 and non >=10 and ref >=13: passflag=1
			elif 'DUP_snv' in clfname and non >=10 and ref >=10 and 'DUP_breakpoint' in self.clf: passflag=1
			elif 'DUP_snv' in clfname and 'DUP_breakpoint' not in self.clf:
				if  svlen<5000 and ref >=18 and non >=18: passflag=1
				elif 5000<= svlen<100000 and ref >=15 and non >=10: passflag=1
				elif svlen>=100000 and ref >=13 and non >=10: passflag=1
			elif 'DEL_male_sex_chrom' in clfname and non>=8 and ref>=8: passflag=1
			elif 'DUP_male_sex_chrom' in clfname and  svlen>=5000 and ref>=10 and non>=10: passflag=1
			if passflag==0: nay+=1
			else: yea+=1
		if nay>yea: self.denovo_filter='FAIL'
	def add_standard_filters(self):
		yea,nay=0,0
		for clfname in self.alt:
			lik,feat,svlen = self.alt[clfname],self.breakpoint_feats,self.locus[2]-self.locus[1]
			passflag=1
			if 'DEL' in clfname and 'male_sex_chrom' not in clfname:
				if feat!=0 and lik<8: passflag=0
				elif feat==0:
					if 'DEL_lt1kb' in clfname and lik <18: passflag=0
					if 'DEL_gt1kb' in clfname:
						if  svlen < 3000: passflag=0
						elif 3000<= svlen<5000 and lik <20: passflag=0
						elif  svlen >=5000 and lik <18: passflag=0
			elif 'DUP_breakpoint' in clfname:
				if feat != 0:
					if  svlen < 1000 and lik < 12: passflag=0
					if  svlen >= 1000 and lik < 10: passflag=0
				elif feat==0:
					if  svlen < 3000: passflag=0
					elif  svlen >=3000 and lik < 12: passflag=0
			elif 'DUP_snv' in clfname:
				if 'DUP_breakpoint' in self.clf:
					if lik < 8: passflag=0
					if  svlen<3000 and lik < 18: passflag=0
					if  svlen>=3000 and lik < 13: passflag=0
			elif 'DEL_male_sex_chrom' in clfname:
				if lik<8 and feat!=0: passflag=0
				if lik<10 and feat==0 and  svlen<=1000: passflag=0
				if  svlen<1000 and feat==0: passflag=0
			elif 'DUP_male_sex_chrom' in clfname:
				if  svlen<5000: passflag=0
				elif  svlen>=5000 and lik<10: passflag=0
				elif feat==0: passflag=0
			if passflag==1: yea+=1
			else: nay+=1
		if nay>yea: self.standard_filter='FAIL'
	def genotype(self,data=None,label=None,copy_number=None,lik=None,biallele=True):
		self.breakpoint_feats+=float(data[7])+float(data[8])
		if biallele==True:
			gq=','.join((format(float(data[14]),'.2f'),format(float(data[15]),'.2f'),format(float(data[16]),'.2f')))
		else:
			gq=','.join((format(float(data[14]),'.2f'),format(float(data[15]),'.2f')))
		_format = '{}:{}:{}'.format(label,copy_number,':'.join(map(str,(format(float(data[7]),'.2f',), format(float(data[8]),'.2f',), format(float(data[9]),'.2f',), int(data[10]), format(float(data[11]),'.2f'), int(np.rint(float(data[12]))),int(np.rint(lik)),gq))))		
		self.gt[self.locus+(str(data[5]),)]=_format
	def median_likelihood(self):
		if len(self.ref)>0:
			self.med_ref=[]
			for clfname in self.ref:
				self.med_ref+=self.ref[clfname]
				self.ref[clfname]=int(np.rint(np.median(self.ref[clfname])))
			self.med_ref=int(np.rint(np.median(self.med_ref)))
		if len(self.alt)>0:
			self.med_alt=[]
			for clfname in self.alt:
				self.med_alt+=self.alt[clfname]
				self.alt[clfname]=int(np.rint(np.median(self.alt[clfname])))
			self.med_alt=int(np.rint(np.median(self.med_alt)))


#!/usr/bin/env python
from sv2_backend import format_chrom,reciprocal_overlap
from sv2Config import Config
import os,pybedtools
from pybedtools import BedTool
class Annotation():
	def __init__(self):
		self.cytoband={} # [(c,s,e)]=cytoband
		self.excluded={}
		self.repeatmasker={} #[(c,s,e)]= id,ovr
		self._1kgp={} #[(c,s,e,cl)]= id,ovr
		self.genes={}
	def check_overlap(self,svs=None,raw=None,gen=None,tmp_dir=None):
		# overlap cytobands
		try:
			for entry in svs.intersect('{}annotation_files/{}_cytoband.bed'.format(Config().resource_path(),gen),wao=True):
				locus = tokenize_sv(entry)
				if self.cytoband.get(locus)==None: self.cytoband[locus]=str(entry[3]).replace('chr','')+str(entry[6])
				else: self.cytoband[locus]=self.cytoband[locus]+','+str(entry[3]).replace('chr','')+str(entry[6])
		except pybedtools.cbedtools.MalformedBedLineError: self.cytoband[locus]='NA'
		# overlap excluded elements
		tmp_bed=tmp_dir+'tmp_anno.bed'
		svs.intersect('{}annotation_files/{}_excluded.bed.gz'.format(Config().resource_path(),gen),wao=True,output=tmp_bed)
		with open(tmp_bed,'r') as f:
			for l in f:
				entry = tuple(l.rstrip().split('\t'))
				locus, ovr = tokenize_sv(entry),int(entry[-1])
				if ovr==0: continue
				key = locus+(str(entry[6]),)
				if self.excluded.get(key)==None: self.excluded[key]=ovr
				else: self.excluded[key]+=ovr
		for locus in self.excluded: self.excluded[locus]=float(self.excluded[locus])/(int(locus[2])-int(locus[1]))
		os.remove(tmp_bed)
		# overlap repeatmasker
		svs.intersect('{}annotation_files/{}_repeatmasker.bed.gz'.format(Config().resource_path(),gen), f=0.8, F=0.8, wa=True, wb=True,output=tmp_bed)
		with open(tmp_bed,'r') as f:
			for l in f:
				entry = tuple(l.rstrip().split('\t'))
				locus, rovr=tokenize_sv(entry), reciprocal_overlap(map(int,(entry[1],entry[4],entry[2],entry[5])))
				name = '{}:{}:{}'.format(entry[7],entry[8],entry[6])
				if self.repeatmasker.get(locus)==None:
					self.repeatmasker[locus]=(name, rovr)
				elif self.repeatmasker.get(locus)!=None and rovr > self.repeatmasker[locus][1]:
					self.repeatmasker[locus]=(name, rovr)
				else: continue
		os.remove(tmp_bed)
		# overlap 1KGP phase 3 DEL/DUP
		if 'hg' in gen:
			self.load_1kgp(raw,'DEL',gen,tmp_bed)
			self.load_1kgp(raw,'DUP',gen,tmp_bed)
		# overlap genes
		genes={}
		svs.intersect('{}annotation_files/{}_genes.bed.gz'.format(Config().resource_path(),gen), wa=True,wb=True,output=tmp_bed)
		with open(tmp_bed,'r') as f:
			for l in f:
				entry = tuple(l.rstrip().split('\t'))
				locus, gene = tokenize_sv(entry), str(entry[6])
				if genes.get(locus)==None: genes[locus]=[gene]
				elif gene not in genes[locus]: genes[locus].append(gene)
				else: continue
		os.remove(tmp_bed)
		for x in genes:
			ori,exons,introns,trx,genlist = {},{},{},{},[]
			exon_total,exon_num, intron_total,intron_num,exoncnt,exontot,introncnt,introntot=0,0,0,0,0,0,0,0
			for y in genes[x]:
				gene = y.split(',')
				ori[gene[0]]=gene[1]
				if 'exon' in y:
					trx[gene[0]]=gene[2]
					exon_total = int(gene[len(gene)-1].split('/').pop())
					exon_num = int(gene[len(gene)-1].split('/').pop(0).replace('exon_',''))
					exons[(gene[0],'TOTAL')]=exon_total
					if exons.get(gene[0])==None:
						exons[gene[0]]=1
						exons[(gene[0],exon_num)]=1
					elif exons.get((gene[0],exon_num)) == None:
						exons[gene[0]]+=1
						exons[(gene[0],exon_num)]=1
					else: continue
				elif 'UTR3' in y:
					trx[gene[0]]=gene[2]
					exons[(gene[0],'UTR3')]=1
				elif 'UTR5' in y:
					trx[gene[0]]=gene[2]
					exons[(gene[0],'UTR5')]=1
				elif 'intron' in y:
					trx[gene[0]]=gene[2]
					intron_total = int(gene[len(gene)-1].split('/').pop())
					intron_num = int(gene[len(gene)-1].split('/').pop(0).replace('intron_',''))
					introns[(gene[0],'TOTAL')]=intron_total
					if introns.get(gene[0])==None:
						introns[gene[0]]=1
						introns[(gene[0],intron_num)]=1
					elif introns.get((gene[0],intron_num)) == None:
						introns[gene[0]]+=1
						introns[(gene[0],intron_num)]=1
					else: continue
				elif 'stream' in y:
					trx[gene[0]]=gene[2]
					exons[(gene[0],gene[3])]=1
				else: continue
			for y in trx:
				if ori.get(y) == None: continue
				orient=ori[y]
				exoncnt=0
				exontot=0
				if exons.get(y) != None: exoncnt=exons[y]
				if exons.get((y,'TOTAL')) != None: exontot=exons[(y,'TOTAL')]
				if orient == '+':
					if exons.get((y,'upstream_1kb')) == 1 and exons.get((y,'downstream_1kb')) != 1:
						genlist.append(','.join(map(str,(y,trx[y],'upstream_1kb'))))
					if exons.get((y,'UTR3'))!=1 and exons.get((y,'UTR5')) == 1:
						genlist.append(','.join(map(str,(y,trx[y],'UTR5'))))
					if exoncnt != 0: genlist.append(','.join(map(str,(y,trx[y],'exon_{}/{}'.format(exoncnt,exontot)))))
					if exons.get(y) == None or exoncnt != exontot:
						introncnt,introntot=0,0
						if introns.get(y) != None: introncnt=introns[y]
						if introns.get((y,'TOTAL'))!=None: introntot=introns[(y,'TOTAL')]
						if introncnt !=0: genlist.append(','.join(map(str,(y,trx[y],'intron_{}/{}'.format(introncnt,introntot)))))
					if exons.get((y,'UTR3'))==1 and exons.get((y,'UTR5')) != 1:
						genlist.append(','.join(map(str,(y,trx[y],'UTR3'))))
					if exons.get((y,'upstream_1kb')) != 1 and exons.get((y,'downstream_1kb')) == 1:
						genlist.append(','.join(map(str,(y,trx[y],'downstream_1kb'))))
				else:
					if exons.get((y,'upstream_1kb')) != 1 and exons.get((y,'downstream_1kb')) == 1:
						genlist.append(','.join(map(str,(y,trx[y],'downstream_1kb'))))
					if exons.get((y,'UTR3'))==1 and exons.get((y,'UTR5')) != 1:
						genlist.append(','.join(map(str,(y,trx[y],'UTR3'))))
					if exoncnt != 0: genlist.append(','.join(map(str,(y,trx[y],'exon_{}/{}'.format(exoncnt,exontot)))))
					if exons.get(y) == None or exoncnt != exontot:
						introncnt,introntot=0,0
						if introns.get(y) != None: introncnt=introns[y]
						if introns.get((y,'TOTAL'))!=None: introntot=introns[(y,'TOTAL')]
						if introncnt !=0: genlist.append(','.join(map(str,(y,trx[y],'intron_{}/{}'.format(introncnt,introntot)))))
					if exons.get((y,'UTR3'))!=1 and exons.get((y,'UTR5')) == 1:
						genlist.append(','.join(map(str,(y,trx[y],'UTR5'))))
					if exons.get((y,'upstream_1kb')) == 1 and exons.get((y,'downstream_1kb')) != 1:
						genlist.append(','.join(map(str,(y,trx[y],'upstream_1kb'))))
			if len(genlist)>=1 : self.genes[x]='|'.join(genlist)
	def load_1kgp(self,raw=None,svtype=None,gen=None,tmp_bed=None):
		sv = BedTool([(format_chrom(x[0]),x[1],x[2],x[3]) for x in raw if svtype in str(x[3])]).sort()
		sv.intersect('{}annotation_files/{}_1000Genomes_{}.bed'.format(Config().resource_path(),gen,svtype), f=0.8, F=0.8, wao=True,output=tmp_bed)
		with open(tmp_bed,'r') as f:
			for l in f:
				x = tuple(l.rstrip().split('\t'))
				locus = tokenize_sv(x)+(str(x[3]),)
				ovr = int(x[-1])
				if ovr==0: continue
				ovr = format(float(x[len(x)-1])/(int(x[2])-int(x[1])),'.2f')
				if self._1kgp.get(locus)==None:
					self._1kgp[locus]=(x[len(x)-2],ovr)
				elif self._1kgp.get(locus)!=None and float(ovr) > float(self._1kgp[locus][1]):
					self._1kgp[locus]=(x[len(x)-2],ovr)
				else: continue
		os.remove(tmp_bed)
def tokenize_sv(x): return (format_chrom(str(x[0])),int(x[1]),int(x[2]))


import os
import numpy
import sys
import glob
import fnmatch
from Cython.Build import cythonize
#from distutils.core import setup, Extension
from setuptools import setup,Extension
import distutils.command.install_data
def opj(*args):
        path = os.path.join(*args)
        return os.path.normpath(path)
class wx_smart_install_data(distutils.command.install_data.install_data):
        def run(self):
                install_cmd = self.get_finalized_command('install')
                self.install_dir = getattr(install_cmd, 'install_lib')
                return distutils.command.install_data.install_data.run(self)
def find_data_files(srcdir, *wildcards, **kw):
        OMIT=['.c','.pyc','.egg-info','.so']
        def walk_helper(arg, dirname, files):
                if '.svn' in dirname or 'CVS' in dirname:
                        return
                names = []
                lst, wildcards = arg
                for wc in wildcards:
                        wc_name = opj(dirname, wc)
                        for f in files:
                                filename = opj(dirname, f)
                                if not any(filename.endswith(x) for x in OMIT):
                                        if fnmatch.fnmatch(filename, wc_name) and not os.path.isdir(filename):
                                                names.append(filename)
                if names:
                        lst.append( (dirname, names ) )
        file_list = []
        recursive = kw.get('recursive', True)
        if recursive:
                os.path.walk(srcdir, walk_helper, (file_list, wildcards))
        else:
                walk_helper((file_list, wildcards),
                                        srcdir,
                                        [os.path.basename(f) for f in glob.glob(opj(srcdir, '*'))])
        return file_list
files = find_data_files('sv2/', '*.*')
setup(  name='sv2',
        version='1.4.3.2',
        description='SV2: Support Vector Structural Variation Genotyper',
        url='https://github.com/dantaki/SV2',
        author='Danny Antaki',
        author_email='dantaki@ucsd.edu',
        license='MIT',
        packages=['sv2'],
        package_dir={'sv2': 'sv2/'},
        package_data={ 
            'sv2': ['sv2/resources/*','sv2/resources/training_sets/*',]
            },
        ext_modules=cythonize([
            #Extension('sv2.core',['sv2/core.pyx']),]
            Extension('sv2.FeatureExtraction',['sv2/FeatureExtraction.pyx']),
            Extension('sv2.Genotype',['sv2/Genotype.pyx']),
            Extension('sv2.Preprocess',['sv2/Preprocess.pyx']),
            Extension('sv2.Snv',['sv2/Snv.pyx']),
            Extension('sv2.Svm',['sv2/Svm.pyx']),
            ]
            
        ),
	include_dirs=[numpy.get_include()],
        requires=[
                'cython',
                'json',
		'numpy',
                'pandas',
                'pybedtools',
                'pysam',
		'shutil',
                'wget',
	],
	python_requires='~=2.7',
        include_package_data=True,
	install_requires= [
                'cython',
                'numpy',
		'pandas',
		'pybedtools',
		'pysam>=0.9',
		'scikit-learn>=0.19',
                'wget',
	],
	data_files = files,
	cmdclass = {'install_data': wx_smart_install_data},
        scripts = ['sv2/sv2','sv2/sv2train']
         )



#!/usr/bin/env python
from sv2_backend import errFH,tokenize
import sys
class Ped():
	def __init__(self,peds=None):
		self.sex={}
		self.males={}
		self.ids=[]
		for fh in peds:
			errFH(fh)
			with open(fh,'r') as f:
				for l in f:
					r = tokenize(l)
					if r!=0:
						if len(r)<5: sys.stderr.write('WARNING {} does not contain 5 elements. PED files are formatted as:\nFamily ID  Individual ID  Paternal ID  Maternal ID  Sex(1=male;2=female;)\n'.format(l))
						else:
							sample_id,sex = str(r[1]),int(r[4])
							if sex != 1 and sex != 2: sys.stderr.write('WARNING {} is not an accepted sex entry. Accepted entries: 1=male; 2=female. Error found here:{}\n'.format(sex,l))
							else:
								self.sex[sample_id]=sex
								self.ids.append(sample_id)
								if sex==1: self.males[sample_id]=1
		self.ids=sorted(list(set(self.ids)),key=str.lower)
def ped_init(peds=None): return Ped(peds)